from __future__ import unicode_literals

import re
from collections import deque
from typing import Callable, Deque, Dict, List, Tuple

import six
from prompt_toolkit.auto_suggest import DynamicAutoSuggest
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion import DynamicCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.filters import Condition, has_focus, is_done, is_true, to_filter
from prompt_toolkit.formatted_text.base import StyleAndTextTuples
from prompt_toolkit.layout.containers import Window, WindowAlign
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.layout.margins import NumberedMargin, ScrollbarMargin
from prompt_toolkit.layout.processors import AppendAutoSuggestion, BeforeInput, ConditionalProcessor, PasswordProcessor
from prompt_toolkit.lexers import DynamicLexer
from prompt_toolkit.lexers.base import Lexer
from prompt_toolkit.widgets.toolbars import SearchToolbar

from hummingbot.client.config.config_helpers import ClientConfigAdapter
from hummingbot.client.ui.style import load_style, text_ui_style


class CustomBuffer(Buffer):
    def validate_and_handle(self):
        valid = self.validate(set_cursor=True)
        if valid:
            if self.accept_handler:
                keep_text = self.accept_handler(self)
            else:
                keep_text = False
            if not keep_text:
                self.reset()


class FormattedTextLexer(Lexer):

    PROMPT_TEXT = ">>> "

    def __init__(self, client_config_map: ClientConfigAdapter) -> None:
        super().__init__()
        self.html_tag_css_style_map: Dict[str, str] = {
            style: css for style, css in load_style(client_config_map).style_rules
        }
        self.html_tag_css_style_map.update({
            ti.attr: ti.value
            for ti in client_config_map.color.traverse()
            if ti.attr not in self.html_tag_css_style_map
        })

        # Maps specific text to its corresponding UI styles
        self.text_style_tag_map: Dict[str, str] = text_ui_style

    def get_css_style(self, tag: str) -> str:
        style = self.html_tag_css_style_map.get(tag, "")
        return style

    def lex_document(self, document: Document) -> Callable[[int], StyleAndTextTuples]:
        lines = document.lines

        def get_line(lineno: int) -> StyleAndTextTuples:
            "Return the tokens for the given line."
            try:
                current_line = lines[lineno]

                # Apply styling to command prompt
                if current_line.startswith(self.PROMPT_TEXT):
                    return [(self.get_css_style("primary_label"), current_line)]

                matched_indexes: List[Tuple[int, int, str]] = [(match.start(), match.end(), style)
                                                               for special_word, style in self.text_style_tag_map.items()
                                                               for match in list(re.finditer(special_word, current_line))
                                                               ]
                if len(matched_indexes) == 0:
                    return [("", current_line)]

                previous_idx = 0
                line_fragments = []
                for start_idx, end_idx, style in matched_indexes:
                    line_fragments.extend([
                        ("", current_line[previous_idx:start_idx]),
                        (self.get_css_style("output_pane"), current_line[start_idx:start_idx + 2]),
                        (self.get_css_style(style), current_line[start_idx + 2:end_idx])
                    ])
                    previous_idx = end_idx

                line_fragments.append(("", current_line[previous_idx:]))

                return line_fragments
            except IndexError:
                return []

        return get_line


class CustomTextArea:
    def __init__(self, text='', multiline=True, password=False,
                 lexer=None, auto_suggest=None, completer=None,
                 complete_while_typing=True, accept_handler=None, history=None,
                 focusable=True, focus_on_click=False, wrap_lines=True,
                 read_only=False, width=None, height=None,
                 dont_extend_height=False, dont_extend_width=False,
                 line_numbers=False, get_line_prefix=None, scrollbar=False,
                 style='', search_field=None, preview_search=True, prompt='',
                 input_processors=None, max_line_count=1000, initial_text="", align=WindowAlign.LEFT):
        assert isinstance(text, six.text_type)
        assert search_field is None or isinstance(search_field, SearchToolbar)

        if search_field is None:
            search_control = None
        elif isinstance(search_field, SearchToolbar):
            search_control = search_field.control

        if input_processors is None:
            input_processors = []

        # Writeable attributes.
        self.completer = completer
        self.complete_while_typing = complete_while_typing
        self.lexer = lexer
        self.auto_suggest = auto_suggest
        self.read_only = read_only
        self.wrap_lines = wrap_lines
        self.max_line_count = max_line_count

        self.buffer = CustomBuffer(
            document=Document(text, 0),
            multiline=multiline,
            read_only=Condition(lambda: is_true(self.read_only)),
            completer=DynamicCompleter(lambda: self.completer),
            complete_while_typing=Condition(
                lambda: is_true(self.complete_while_typing)),
            auto_suggest=DynamicAutoSuggest(lambda: self.auto_suggest),
            accept_handler=accept_handler,
            history=history)

        self.control = BufferControl(
            buffer=self.buffer,
            lexer=DynamicLexer(lambda: self.lexer),
            input_processors=[
                ConditionalProcessor(
                    AppendAutoSuggestion(),
                    has_focus(self.buffer) & ~is_done),
                ConditionalProcessor(
                    processor=PasswordProcessor(),
                    filter=to_filter(password)
                ),
                BeforeInput(prompt, style='class:text-area.prompt'),
            ] + input_processors,
            search_buffer_control=search_control,
            preview_search=preview_search,
            focusable=focusable,
            focus_on_click=focus_on_click)

        if multiline:
            if scrollbar:
                right_margins = [ScrollbarMargin(display_arrows=True)]
            else:
                right_margins = []
            if line_numbers:
                left_margins = [NumberedMargin()]
            else:
                left_margins = []
        else:
            left_margins = []
            right_margins = []

        style = 'class:text-area ' + style

        self.window = Window(
            height=height,
            width=width,
            dont_extend_height=dont_extend_height,
            dont_extend_width=dont_extend_width,
            content=self.control,
            style=style,
            wrap_lines=Condition(lambda: is_true(self.wrap_lines)),
            left_margins=left_margins,
            right_margins=right_margins,
            get_line_prefix=get_line_prefix,
            align=align)

        self.log_lines: Deque[str] = deque()
        self.log(initial_text)

    @property
    def text(self):
        """
        The `Buffer` text.
        """
        return self.buffer.text

    @text.setter
    def text(self, value):
        self.buffer.set_document(Document(value, 0), bypass_readonly=True)

    @property
    def document(self):
        """
        The `Buffer` document (text + cursor position).
        """
        return self.buffer.document

    @document.setter
    def document(self, value):
        self.buffer.document = value

    @property
    def accept_handler(self):
        """
        The accept handler. Called when the user accepts the input.
        """
        return self.buffer.accept_handler

    @accept_handler.setter
    def accept_handler(self, value):
        self.buffer.accept_handler = value

    def __pt_container__(self):
        return self.window

    def log(self, text: str, save_log: bool = True, silent: bool = False):
        # Getting the max width of the window area
        if self.window.render_info is None:
            max_width = 100
        else:
            max_width = self.window.render_info.window_width - 2

        # remove simple formatting tags
        repls = (('<b>', ''), ('</b>', ''), ('<pre>', ''), ('</pre>', ''))
        for r in repls:
            text = text.replace(*r)

        # Split the string into multiple lines if there is a "\n" or if the string exceeds max window width
        # This operation should not be too expensive because only the newly added lines are processed
        new_lines_raw: List[str] = str(text).split('\n')
        new_lines = []
        for line in new_lines_raw:
            while len(line) > max_width:
                new_lines.append(line[0:max_width])
                line = line[max_width:]
            new_lines.append(line)

        if save_log:
            self.log_lines.extend(new_lines)
            while len(self.log_lines) > self.max_line_count:
                self.log_lines.popleft()
            new_text: str = "\n".join(self.log_lines)
        else:
            new_text: str = "\n".join(new_lines)
        if not silent:
            self.buffer.document = Document(text=new_text, cursor_position=len(new_text))



from pydantic import Field

from hummingbot.client.config.config_data_types import BaseClientModel


class SSLConfigMap(BaseClientModel):
    caCertificatePath: str = Field(default="/usr/src/app/certs/ca_cert.pem")
    certificatePath: str = Field(default="/usr/src/app/certs/server_cert.pem")
    keyPath: str = Field(default="/usr/src/app/certs/server_key.pem")



from decimal import Decimal
from typing import TYPE_CHECKING, Dict, Optional

from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.rate_oracle.sources.rate_source_base import RateSourceBase
from hummingbot.core.utils import async_ttl_cache

if TYPE_CHECKING:
    from hummingbot.connector.exchange.hyperliquid.hyperliquid_exchange import HyperliquidExchange


class HyperliquidRateSource(RateSourceBase):
    def __init__(self):
        super().__init__()
        self._exchange: Optional[HyperliquidExchange] = None  # delayed because of circular reference

    @property
    def name(self) -> str:
        return "hyperliquid"

    @async_ttl_cache(ttl=30, maxsize=1)
    async def get_prices(self, quote_token: Optional[str] = None) -> Dict[str, Decimal]:
        self._ensure_exchange()
        results = {}
        try:
            pairs_prices = await self._exchange.get_all_pairs_prices()
            for pair_price in pairs_prices:
                pair = pair_price["symbol"]
                try:
                    trading_pair = await self._exchange.trading_pair_associated_to_exchange_symbol(symbol=pair)
                except KeyError:
                    continue  # skip pairs that we don't track
                if quote_token is not None:
                    base, quote = split_hb_trading_pair(trading_pair=trading_pair)
                    if quote != quote_token:
                        continue
                price = pair_price["price"]
                if price is not None:
                    results[trading_pair] = Decimal(price)
        except Exception:
            self.logger().exception(
                msg="Unexpected error while retrieving rates from Hyperliquid. Check the log file for more info.",
            )
        return results

    def _ensure_exchange(self):
        if self._exchange is None:
            self._exchange = self._build_hyperliquid_connector_without_private_keys()

    @staticmethod
    def _build_hyperliquid_connector_without_private_keys() -> 'HyperliquidExchange':
        from hummingbot.connector.exchange.hyperliquid.hyperliquid_exchange import HyperliquidExchange

        return HyperliquidExchange(
            hyperliquid_secret_key="",
            trading_pairs=[],
            use_vault=False,
            hyperliquid_mode = "arb_wallet",
            hyperliquid_address="",
            trading_required=False,
        )



import logging
import time
from typing import List, Optional

import pandas as pd

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.tracking_nonce import get_tracking_nonce
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.data_feed.candles_feed.kucoin_spot_candles import constants as CONSTANTS
from hummingbot.logger import HummingbotLogger


class KucoinSpotCandles(CandlesBase):
    _logger: Optional[HummingbotLogger] = None
    _last_ws_message_sent_timestamp = 0
    _ping_interval = 0

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 150):
        super().__init__(trading_pair, interval, max_records)
        self._ws_url = None
        self._ws_token = None

    @property
    def name(self):
        return f"kucoin_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        return f"{self._ws_url}?token={self._ws_token}"

    @property
    def health_check_url(self):
        return self.rest_url + CONSTANTS.HEALTH_CHECK_ENDPOINT

    @property
    def candles_url(self):
        return self.rest_url + CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def public_ws_url(self):
        return self.rest_url + CONSTANTS.PUBLIC_WS_DATA_PATH_URL

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    @property
    def candles_df(self) -> pd.DataFrame:
        df = pd.DataFrame(self._candles, columns=self.columns, dtype=float)
        return df.sort_values(by="timestamp", ascending=True)

    @property
    def _ping_payload(self):
        return {
            "type": "ping"
        }

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        await rest_assistant.execute_request(url=self.health_check_url,
                                             throttler_limit_id=CONSTANTS.HEALTH_CHECK_ENDPOINT)
        return NetworkStatus.CONNECTED

    def get_exchange_trading_pair(self, trading_pair):
        return trading_pair

    @property
    def _is_last_candle_not_included_in_rest_request(self):
        return True

    @property
    def _is_first_candle_not_included_in_rest_request(self):
        return False

    def _get_rest_candles_params(self,
                                 start_time: Optional[int] = None,
                                 end_time: Optional[int] = None,
                                 limit: Optional[int] = CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST) -> dict:
        """
        For API documentation, please refer to:
        https://www.kucoin.com/docs/rest/spot-trading/market-data/get-klines
        """
        params = {"symbol": self._ex_trading_pair, "type": CONSTANTS.INTERVALS[self.interval]}
        if start_time:
            params["startAt"] = start_time
        if end_time:
            params["endAt"] = end_time
        return params

    def _parse_rest_candles(self, data: dict, end_time: Optional[int] = None) -> List[List[float]]:
        return [[self.ensure_timestamp_in_seconds(row[0]), row[1], row[3], row[4], row[2], row[5], row[6], 0., 0., 0.]
                for row in data['data']][::-1]

    def ws_subscription_payload(self):
        return {
            "id": str(get_tracking_nonce()),
            "type": "subscribe",
            "topic": f"/market/candles:{self._ex_trading_pair}_{CONSTANTS.INTERVALS[self.interval]}",
            "privateChannel": False,
            "response": False,
        }

    def _parse_websocket_message(self, data: dict):
        candles_row_dict = {}
        if data is not None and data.get(
                "subject") == "trade.candles.update":  # data will be None when the websocket is disconnected
            candles = data["data"]["candles"]
            candles_row_dict["timestamp"] = self.ensure_timestamp_in_seconds(candles[0])
            candles_row_dict["open"] = candles[1]
            candles_row_dict["close"] = candles[2]
            candles_row_dict["high"] = candles[3]
            candles_row_dict["low"] = candles[4]
            candles_row_dict["volume"] = candles[5]
            candles_row_dict["quote_asset_volume"] = candles[6]
            candles_row_dict["n_trades"] = 0.
            candles_row_dict["taker_buy_base_volume"] = 0.
            candles_row_dict["taker_buy_quote_volume"] = 0.
            return candles_row_dict

    async def initialize_exchange_data(self):
        rest_assistant = await self._api_factory.get_rest_assistant()
        connection_info = await rest_assistant.execute_request(
            url=self.public_ws_url,
            method=RESTMethod.POST,
            throttler_limit_id=CONSTANTS.PUBLIC_WS_DATA_PATH_URL,
        )

        self._ws_url = connection_info["data"]["instanceServers"][0]["endpoint"]
        self._ping_timeout = int(connection_info["data"]["instanceServers"][0]["pingTimeout"]) * 1e-3
        self._ws_token = connection_info["data"]["token"]

    @staticmethod
    def _time():
        return time.time()



import asyncio
from decimal import Decimal
from typing import Dict, List, Iterator, Mapping, Optional

from bidict import bidict

from hummingbot.connector.budget_checker import BudgetChecker
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.core.data_type.common import OrderType, PriceType, TradeType
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_query_result import ClientOrderBookQueryResult, OrderBookQueryResult
from hummingbot.core.data_type.order_book_row import ClientOrderBookRow
from hummingbot.core.data_type.order_book_tracker import OrderBookTracker
from hummingbot.core.data_type.trade_fee import AddedToCostTradeFee
from hummingbot.core.utils.async_utils import safe_gather


s_float_NaN = float("nan")
s_decimal_NaN = Decimal("nan")
s_decimal_0 = Decimal(0)


cdef class ExchangeBase(ConnectorBase):
    """
    ExchangeBase provides common exchange (for both centralized and decentralized) connector functionality and
    interface.
    """

    def __init__(self,
                 balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
                 rate_limits_share_pct: Decimal = Decimal("100")):
        super().__init__(balance_asset_limit)
        self._order_book_tracker = None
        self._budget_checker = BudgetChecker(exchange=self)
        self._trading_pair_symbol_map: Optional[Mapping[str, str]] = None
        self._mapping_initialization_lock = asyncio.Lock()

    @staticmethod
    def convert_from_exchange_trading_pair(exchange_trading_pair: str) -> Optional[str]:
        return exchange_trading_pair

    @staticmethod
    def convert_to_exchange_trading_pair(hb_trading_pair: str) -> str:
        return hb_trading_pair

    @property
    def order_books(self) -> Dict[str, OrderBook]:
        raise NotImplementedError

    @property
    def limit_orders(self) -> List[LimitOrder]:
        raise NotImplementedError

    @property
    def budget_checker(self) -> BudgetChecker:
        return self._budget_checker

    @property
    def order_book_tracker(self) -> Optional[OrderBookTracker]:
        return self._order_book_tracker

    async def trading_pair_symbol_map(self):
        if not self.trading_pair_symbol_map_ready():
            async with self._mapping_initialization_lock:
                if not self.trading_pair_symbol_map_ready():
                    await self._initialize_trading_pair_symbol_map()
        current_map = self._trading_pair_symbol_map or bidict()
        return current_map

    def trading_pair_symbol_map_ready(self):
        """
        Checks if the mapping from exchange symbols to client trading pairs has been initialized

        :return: True if the mapping has been initialized, False otherwise
        """
        return self._trading_pair_symbol_map is not None and len(self._trading_pair_symbol_map) > 0

    async def all_trading_pairs(self) -> List[str]:
        """
        List of all trading pairs supported by the connector

        :return: List of trading pair symbols in the Hummingbot format
        """
        mapping = await self.trading_pair_symbol_map()
        return list(mapping.values())

    async def exchange_symbol_associated_to_pair(self, trading_pair: str) -> str:
        """
        Used to translate a trading pair from the client notation to the exchange notation

        :param trading_pair: trading pair in client notation

        :return: trading pair in exchange notation
        """
        symbol_map = await self.trading_pair_symbol_map()
        return symbol_map.inverse[trading_pair]

    async def trading_pair_associated_to_exchange_symbol(self, symbol: str,) -> str:
        """
        Used to translate a trading pair from the exchange notation to the client notation

        :param symbol: trading pair in exchange notation

        :return: trading pair in client notation
        """
        symbol_map = await self.trading_pair_symbol_map()
        return symbol_map[symbol]

    async def get_last_traded_prices(self, trading_pairs: List[str]) -> Dict[str, float]:
        """
        Return a dictionary the trading_pair as key and the current price as value for each trading pair passed as
        parameter

        :param trading_pairs: list of trading pairs to get the prices for

        :return: Dictionary of associations between token pair and its latest price
        """
        tasks = [self._get_last_traded_price(trading_pair=trading_pair) for trading_pair in trading_pairs]
        results = await safe_gather(*tasks)
        return {t_pair: result for t_pair, result in zip(trading_pairs, results)}

    def get_mid_price(self, trading_pair: str) -> Decimal:
        return (self.get_price(trading_pair, True) + self.get_price(trading_pair, False)) / Decimal("2")

    cdef str c_buy(self, str trading_pair, object amount, object order_type=OrderType.MARKET,
                   object price=s_decimal_NaN, dict kwargs={}):
        return self.buy(trading_pair, amount, order_type, price, **kwargs)

    cdef str c_sell(self, str trading_pair, object amount, object order_type=OrderType.MARKET,
                    object price=s_decimal_NaN, dict kwargs={}):
        return self.sell(trading_pair, amount, order_type, price, **kwargs)

    cdef c_cancel(self, str trading_pair, str client_order_id):
        return self.cancel(trading_pair, client_order_id)

    cdef c_stop_tracking_order(self, str order_id):
        raise NotImplementedError

    cdef object c_get_fee(self,
                          str base_currency,
                          str quote_currency,
                          object order_type,
                          object order_side,
                          object amount,
                          object price,
                          object is_maker = None):
        return self.get_fee(base_currency, quote_currency, order_type, order_side, amount, price, is_maker)

    cdef OrderBook c_get_order_book(self, str trading_pair):
        return self.get_order_book(trading_pair)

    cdef object c_get_price(self, str trading_pair, bint is_buy):
        """
        :returns: Top bid/ask price for a specific trading pair
        """
        cdef:
            OrderBook order_book = self.c_get_order_book(trading_pair)
            object top_price
        try:
            top_price = Decimal(str(order_book.c_get_price(is_buy)))
        except EnvironmentError as e:
            self.logger().warning(f"{'Ask' if is_buy else 'Bid'} orderbook for {trading_pair} is empty.")
            return s_decimal_NaN
        return self.c_quantize_order_price(trading_pair, top_price)

    cdef ClientOrderBookQueryResult c_get_vwap_for_volume(self, str trading_pair, bint is_buy, object volume):
        cdef:
            OrderBook order_book = self.c_get_order_book(trading_pair)
            OrderBookQueryResult result = order_book.c_get_vwap_for_volume(is_buy, float(volume))
            object query_volume = Decimal(str(result.query_volume))
            object result_price = Decimal(str(result.result_price))
            object result_volume = Decimal(str(result.result_volume))
        return ClientOrderBookQueryResult(s_decimal_NaN,
                                          query_volume,
                                          result_price,
                                          result_volume)

    cdef ClientOrderBookQueryResult c_get_price_for_quote_volume(self, str trading_pair, bint is_buy, double volume):
        cdef:
            OrderBook order_book = self.c_get_order_book(trading_pair)
            OrderBookQueryResult result = order_book.c_get_price_for_quote_volume(is_buy, float(volume))
            object query_volume = Decimal(str(result.query_volume))
            object result_price = Decimal(str(result.result_price))
            object result_volume = Decimal(str(result.result_volume))
        return ClientOrderBookQueryResult(s_decimal_NaN,
                                          query_volume,
                                          result_price,
                                          result_volume)

    cdef ClientOrderBookQueryResult c_get_price_for_volume(self, str trading_pair, bint is_buy, object volume):
        cdef:
            OrderBook order_book = self.c_get_order_book(trading_pair)
            OrderBookQueryResult result = order_book.c_get_price_for_volume(is_buy, float(volume))
            object query_volume = Decimal(str(result.query_volume))
            object result_price = Decimal(str(result.result_price))
            object result_volume = Decimal(str(result.result_volume))
        return ClientOrderBookQueryResult(s_decimal_NaN,
                                          query_volume,
                                          result_price,
                                          result_volume)

    cdef ClientOrderBookQueryResult c_get_quote_volume_for_base_amount(self, str trading_pair, bint is_buy,
                                                                       object base_amount):
        cdef:
            OrderBook order_book = self.c_get_order_book(trading_pair)
            OrderBookQueryResult result = order_book.c_get_quote_volume_for_base_amount(is_buy, float(base_amount))
            object query_volume = Decimal(str(result.query_volume))
            object result_volume = Decimal(str(result.result_volume))
        return ClientOrderBookQueryResult(s_decimal_NaN,
                                          query_volume,
                                          s_decimal_NaN,
                                          result_volume)

    cdef ClientOrderBookQueryResult c_get_volume_for_price(self, str trading_pair, bint is_buy, object price):
        cdef:
            OrderBook order_book = self.c_get_order_book(trading_pair)
            OrderBookQueryResult result = order_book.c_get_volume_for_price(is_buy, float(price))
            object query_price = Decimal(str(result.query_price))
            object result_price = Decimal(str(result.result_price))
            object result_volume = Decimal(str(result.result_volume))
        return ClientOrderBookQueryResult(query_price,
                                          s_decimal_NaN,
                                          result_price,
                                          result_volume)

    cdef ClientOrderBookQueryResult c_get_quote_volume_for_price(self, str trading_pair, bint is_buy, object price):
        cdef:
            OrderBook order_book = self.c_get_order_book(trading_pair)
            OrderBookQueryResult result = order_book.c_get_volume_for_price(is_buy, float(price))
            object query_price = Decimal(str(result.query_price))
            object result_price = Decimal(str(result.result_price))
            object result_volume = Decimal(str(result.result_volume))
        return ClientOrderBookQueryResult(query_price,
                                          s_decimal_NaN,
                                          result_price,
                                          result_volume)

    def order_book_bid_entries(self, trading_pair) -> Iterator[ClientOrderBookRow]:
        cdef:
            OrderBook order_book = self.c_get_order_book(trading_pair)
        for entry in order_book.bid_entries():
            yield ClientOrderBookRow(Decimal(str(entry.price)),
                                     Decimal(str(entry.amount)),
                                     entry.update_id)

    def order_book_ask_entries(self, trading_pair) -> Iterator[ClientOrderBookRow]:
        cdef:
            OrderBook order_book = self.c_get_order_book(trading_pair)
        for entry in order_book.ask_entries():
            yield ClientOrderBookRow(Decimal(str(entry.price)),
                                     Decimal(str(entry.amount)),
                                     entry.update_id)

    def get_vwap_for_volume(self, trading_pair: str, is_buy: bool, volume: Decimal):
        return self.c_get_vwap_for_volume(trading_pair, is_buy, volume)

    def get_price_for_quote_volume(self, trading_pair: str, is_buy: bool, volume: Decimal):
        return self.c_get_price_for_quote_volume(trading_pair, is_buy, volume)

    def get_price_for_volume(self, trading_pair: str, is_buy: bool, volume: Decimal):
        return self.c_get_price_for_volume(trading_pair, is_buy, volume)

    def get_quote_volume_for_base_amount(self, trading_pair: str, is_buy: bool,
                                         base_amount: Decimal) -> ClientOrderBookQueryResult:
        return self.c_get_quote_volume_for_base_amount(trading_pair, is_buy, base_amount)

    def get_volume_for_price(self, trading_pair: str, is_buy: bool, price: Decimal) -> ClientOrderBookQueryResult:
        return self.c_get_volume_for_price(trading_pair, is_buy, price)

    def get_quote_volume_for_price(self, trading_pair: str, is_buy: bool, price: Decimal) -> ClientOrderBookQueryResult:
        return self.c_get_quote_volume_for_price(trading_pair, is_buy, price)

    def get_price(self, trading_pair: str, is_buy: bool) -> Decimal:
        return self.c_get_price(trading_pair, is_buy)

    def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
            price: Decimal = s_decimal_NaN, **kwargs) -> str:
        raise NotImplementedError

    def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
             price: Decimal = s_decimal_NaN, **kwargs) -> str:
        raise NotImplementedError

    def cancel(self, trading_pair: str, client_order_id: str):
        raise NotImplementedError

    def get_order_book(self, trading_pair: str) -> OrderBook:
        raise NotImplementedError

    def get_fee(self,
                base_currency: str,
                quote_currency: str,
                order_type: OrderType,
                order_side: TradeType,
                amount: Decimal,
                price: Decimal = s_decimal_NaN,
                is_maker: Optional[bool] = None) -> AddedToCostTradeFee:
        raise NotImplementedError

    def get_order_price_quantum(self, trading_pair: str, price: Decimal) -> Decimal:
        return self.c_get_order_price_quantum(trading_pair, price)

    def get_order_size_quantum(self, trading_pair: str, order_size: Decimal) -> Decimal:
        return self.c_get_order_size_quantum(trading_pair, order_size)

    def quantize_order_price(self, trading_pair: str, price: Decimal) -> Decimal:
        return self.c_quantize_order_price(trading_pair, price)

    def quantize_order_amount(self, trading_pair: str, amount: Decimal) -> Decimal:
        return self.c_quantize_order_amount(trading_pair, amount)

    def supported_order_types(self):
        return [OrderType.LIMIT, OrderType.MARKET]

    def get_maker_order_type(self):
        """
        Return a maker order type depending what order types the connector supports.
        """
        if OrderType.LIMIT_MAKER in self.supported_order_types():
            return OrderType.LIMIT_MAKER
        elif OrderType.LIMIT in self.supported_order_types():
            return OrderType.LIMIT
        else:
            raise Exception("There is no maker order type supported by this exchange.")

    def get_taker_order_type(self):
        """
        Return a taker order type depending what order types the connector supports.
        """
        if OrderType.MARKET in self.supported_order_types():
            return OrderType.MARKET
        elif OrderType.LIMIT in self.supported_order_types():
            return OrderType.LIMIT
        else:
            raise Exception("There is no taker order type supported by this exchange.")

    def get_price_by_type(self, trading_pair: str, price_type: PriceType) -> Decimal:
        """
        Gets price by type (BestBid, BestAsk, MidPrice or LastTrade)
        :param trading_pair: The market trading pair
        :param price_type: The price type
        :returns The price
        """
        if price_type is PriceType.BestBid:
            return self.c_get_price(trading_pair, False)
        elif price_type is PriceType.BestAsk:
            return self.c_get_price(trading_pair, True)
        elif price_type is PriceType.MidPrice:
            return (self.c_get_price(trading_pair, True) + self.c_get_price(trading_pair, False)) / Decimal("2")
        elif price_type is PriceType.LastTrade:
            return Decimal(self.c_get_order_book(trading_pair).last_trade_price)

    async def get_quote_price(self, trading_pair: str, is_buy: bool, amount: Decimal) -> Decimal:
        """
        For an exchange type connector, the quote price is volume weighted average price.
        """
        return Decimal(str(self.get_vwap_for_volume(trading_pair, is_buy, amount).result_price))

    async def get_order_price(self, trading_pair: str, is_buy: bool, amount: Decimal) -> Decimal:
        """
        For an exchange type connector, the price required for order submission is the price of the order book for
        required volume.
        """
        return Decimal(str(self.get_price_for_volume(trading_pair, is_buy, amount).result_price))

    async def _initialize_trading_pair_symbol_map(self):
        raise NotImplementedError

    def _set_trading_pair_symbol_map(self, trading_pair_and_symbol_map: Optional[Mapping[str, str]]):
        """
        Method added to allow the pure Python subclasses to set the value of the map
        """
        self._trading_pair_symbol_map = trading_pair_and_symbol_map

    def _set_order_book_tracker(self, order_book_tracker: Optional[OrderBookTracker]):
        """
        Method added to allow the pure Python subclasses to store the tracker in the instance variable
        """
        self._order_book_tracker = order_book_tracker

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        raise NotImplementedError



import asyncio
import threading
import time
from typing import TYPE_CHECKING

from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.remote_iface.mqtt import MQTTGateway

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


SUBCOMMANDS = ['start', 'stop', 'restart']


class MQTTCommand:
    _mqtt_sleep_rate_connection_check: float = 1.0
    _mqtt_sleep_rate_autostart_retry: float = 10.0

    def mqtt_start(self,  # type: HummingbotApplication
                   timeout: float = 30.0
                   ):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.mqtt_start, timeout)
            return
        safe_ensure_future(self.start_mqtt_async(timeout=timeout),
                           loop=self.ev_loop)

    def mqtt_stop(self,  # type: HummingbotApplication
                  ):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.mqtt_stop)
            return
        safe_ensure_future(self.stop_mqtt_async(),
                           loop=self.ev_loop)

    def mqtt_restart(self,  # type: HummingbotApplication
                     timeout: float = 30.0
                     ):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.mqtt_restart, timeout)
            return
        safe_ensure_future(self.restart_mqtt_async(timeout=timeout),
                           loop=self.ev_loop)

    async def start_mqtt_async(self,  # type: HummingbotApplication
                               timeout: float = 30.0
                               ):
        if self._mqtt is None:
            while True:
                try:
                    start_t = time.time()
                    self.logger().info('Connecting MQTT Bridge...')
                    self._mqtt = MQTTGateway(self)
                    self._mqtt.start()
                    while True:
                        if time.time() - start_t > timeout:
                            raise Exception(
                                f'Connection timed out after {timeout} seconds')
                        if self._mqtt.health:
                            self.logger().info('MQTT Bridge connected with success.')
                            break
                        await asyncio.sleep(self._mqtt_sleep_rate_connection_check)
                    break
                except Exception as e:
                    if self.client_config_map.mqtt_bridge.mqtt_autostart:
                        s = self._mqtt_sleep_rate_autostart_retry
                        self.logger().error(
                            f'Failed to connect MQTT Bridge: {str(e)}. Retrying in {s} seconds.')
                        self.notify(
                            f'MQTT Bridge failed to connect to the broker, retrying in {s} seconds.'
                        )
                    else:
                        self.logger().error(
                            f'Failed to connect MQTT Bridge: {str(e)}')
                        self.notify('MQTT Bridge failed to connect to the broker.')
                    self._mqtt.stop()
                    self._mqtt = None

                    if self.client_config_map.mqtt_bridge.mqtt_autostart:
                        await asyncio.sleep(self._mqtt_sleep_rate_autostart_retry)
                    else:
                        break

        else:
            self.logger().warning("MQTT Bridge is already running!")
            self.notify('MQTT Bridge is already running!')

    async def stop_mqtt_async(self,  # type: HummingbotApplication
                              ):
        if self._mqtt is not None:
            try:
                self._mqtt.stop()
                self._mqtt = None
                self.logger().info("MQTT Bridge disconnected")
                self.notify('MQTT Bridge disconnected')
            except Exception as e:
                self.logger().error(f'Failed to stop MQTT Bridge: {str(e)}')
        else:
            self.logger().error("MQTT is already stopped!")
            self.notify('MQTT Bridge is already stopped!')

    async def restart_mqtt_async(self,  # type: HummingbotApplication
                                 timeout: float = 2.0
                                 ):
        await self.stop_mqtt_async()
        await self.start_mqtt_async(timeout)



import asyncio
import time
from typing import TYPE_CHECKING, List, Optional

from hummingbot.connector.exchange.mexc import mexc_constants as CONSTANTS, mexc_web_utils as web_utils
from hummingbot.connector.exchange.mexc.mexc_auth import MexcAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.mexc.mexc_exchange import MexcExchange


class MexcAPIUserStreamDataSource(UserStreamTrackerDataSource):
    """
    Manages the user stream connection for MEXC exchange, handling listen key lifecycle
    and websocket connection management.
    """
    LISTEN_KEY_KEEP_ALIVE_INTERVAL = 1800  # Recommended to Ping/Update listen key to keep connection alive
    HEARTBEAT_TIME_INTERVAL = 30.0
    LISTEN_KEY_RETRY_INTERVAL = 5.0  # Delay between listen key management iterations
    MAX_RETRIES = 3  # Maximum retries for obtaining a new listen key

    _logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 auth: MexcAuth,
                 trading_pairs: List[str],
                 connector: 'MexcExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        super().__init__()
        self._auth: MexcAuth = auth
        self._current_listen_key = None
        self._domain = domain
        self._api_factory = api_factory

        # Event to signal when listen key is ready for use
        self._listen_key_initialized_event: asyncio.Event = asyncio.Event()
        # Track last successful ping timestamp for refresh scheduling
        self._last_listen_key_ping_ts = None
        # Background task handle for listen key lifecycle management
        self._manage_listen_key_task = None

    async def _ensure_listen_key_task_running(self):
        """
        Ensures the listen key management task is running.

        Creates a new task if none exists or if the previous task has completed.
        This method is idempotent and safe to call multiple times.
        """
        # If task is already running, do nothing
        if self._manage_listen_key_task is not None and not self._manage_listen_key_task.done():
            return

        # Cancel old task if it exists and is done (failed)
        if self._manage_listen_key_task is not None:
            self._manage_listen_key_task.cancel()
            try:
                await self._manage_listen_key_task
            except asyncio.CancelledError:
                pass
            except Exception:
                pass  # Ignore any exception from the failed task

        # Create new task
        self._manage_listen_key_task = safe_ensure_future(self._manage_listen_key_task_loop())

    async def _connected_websocket_assistant(self) -> WSAssistant:
        """
        Creates an instance of WSAssistant connected to the exchange.

        This method ensures the listen key is ready before connecting.
        The connection process follows these steps:
        1. Ensures the listen key management task is running
        2. Waits for a valid listen key to be obtained
        3. Establishes websocket connection with the listen key

        :return: Connected WSAssistant instance
        :raises: Connection errors if websocket fails to connect
        """
        # Make sure the listen key management task is running
        await self._ensure_listen_key_task_running()

        # Wait for the listen key to be initialized
        await self._listen_key_initialized_event.wait()

        # Get a websocket assistant and connect it
        ws = await self._get_ws_assistant()
        url = f"{CONSTANTS.WSS_URL.format(self._domain)}?listenKey={self._current_listen_key}"

        self.logger().info(f"Connecting to user stream with listen key {self._current_listen_key}")
        await ws.connect(ws_url=url, ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        self.logger().info("Successfully connected to user stream")

        return ws

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        """
        Subscribes to order events and balance events.

        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """
        try:

            orders_change_payload = {
                "method": "SUBSCRIPTION",
                "params": [CONSTANTS.USER_ORDERS_ENDPOINT_NAME],
                "id": 1
            }
            subscribe_order_change_request: WSJSONRequest = WSJSONRequest(payload=orders_change_payload)

            trades_payload = {
                "method": "SUBSCRIPTION",
                "params": [CONSTANTS.USER_TRADES_ENDPOINT_NAME],
                "id": 2
            }
            subscribe_trades_request: WSJSONRequest = WSJSONRequest(payload=trades_payload)

            balance_payload = {
                "method": "SUBSCRIPTION",
                "params": [CONSTANTS.USER_BALANCE_ENDPOINT_NAME],
                "id": 3
            }
            subscribe_balance_request: WSJSONRequest = WSJSONRequest(payload=balance_payload)

            await websocket_assistant.send(subscribe_order_change_request)
            await websocket_assistant.send(subscribe_trades_request)
            await websocket_assistant.send(subscribe_balance_request)

            self.logger().info("Subscribed to private order changes and balance updates channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to user streams...")
            raise

    async def _get_listen_key(self, max_retries: int = MAX_RETRIES) -> str:
        """
        Fetches a listen key from the exchange with retries and exponential backoff.

        Implements a robust retry mechanism to handle temporary network issues
        or API errors. The backoff time doubles after each failed attempt.

        :param max_retries: Maximum number of retry attempts (default: MAX_RETRIES)
        :return: Valid listen key string
        :raises IOError: If all retry attempts fail
        """
        retry_count = 0
        backoff_time = 1.0  # Initial backoff: 1 second
        timeout = 5.0

        rest_assistant = await self._api_factory.get_rest_assistant()
        while True:
            try:
                data = await rest_assistant.execute_request(
                    url=web_utils.public_rest_url(path_url=CONSTANTS.MEXC_USER_STREAM_PATH_URL, domain=self._domain),
                    method=RESTMethod.POST,
                    throttler_limit_id=CONSTANTS.MEXC_USER_STREAM_PATH_URL,
                    is_auth_required=True,
                    timeout=timeout,
                )
                return data["listenKey"]
            except asyncio.CancelledError:
                raise
            except Exception as exception:
                retry_count += 1
                if retry_count > max_retries:
                    raise IOError(f"Error fetching user stream listen key after {max_retries} retries. Error: {exception}")

                self.logger().warning(f"Retry {retry_count}/{max_retries} fetching user stream listen key. Error: {exception}")
                await self._sleep(backoff_time)
                backoff_time *= 2  # Exponential backoff: 1s, 2s, 4s...

    async def _ping_listen_key(self) -> bool:
        rest_assistant = await self._api_factory.get_rest_assistant()
        try:
            data = await rest_assistant.execute_request(
                url=web_utils.public_rest_url(path_url=CONSTANTS.MEXC_USER_STREAM_PATH_URL, domain=self._domain),
                params={"listenKey": self._current_listen_key},
                method=RESTMethod.PUT,
                return_err=True,
                throttler_limit_id=CONSTANTS.MEXC_USER_STREAM_PATH_URL,
                is_auth_required=True
            )

            if "code" in data:
                self.logger().warning(f"Failed to refresh the listen key {self._current_listen_key}: {data}")
                return False

        except asyncio.CancelledError:
            raise
        except Exception as exception:
            self.logger().warning(f"Failed to refresh the listen key {self._current_listen_key}: {exception}")
            return False

        return True

    async def _manage_listen_key_task_loop(self):
        """
        Background task that manages the listen key lifecycle.

        This is the core method that ensures continuous connectivity by:
        1. Obtaining a new listen key if none exists or previous one failed
        2. Periodically refreshing the listen key before it expires (30-minute intervals)
        3. Handling errors gracefully and resetting state when necessary

        The task runs indefinitely until cancelled, automatically recovering from errors.
        State is properly cleaned up in the finally block to ensure consistency.
        """
        self.logger().info("Starting listen key management task...")
        try:
            while True:
                try:
                    now = int(time.time())

                    # Initialize listen key if needed (first run or after error)
                    if self._current_listen_key is None:
                        self._current_listen_key = await self._get_listen_key()
                        self._last_listen_key_ping_ts = now
                        self._listen_key_initialized_event.set()
                        self.logger().info(f"Successfully obtained listen key {self._current_listen_key}")

                    # Refresh listen key periodically to prevent expiration
                    if now - self._last_listen_key_ping_ts >= self.LISTEN_KEY_KEEP_ALIVE_INTERVAL:
                        success = await self._ping_listen_key()
                        if success:
                            self.logger().info(f"Successfully refreshed listen key {self._current_listen_key}")
                            self._last_listen_key_ping_ts = now
                        else:
                            # Ping failed - force obtaining a new key in next iteration
                            self.logger().error(f"Failed to refresh listen key {self._current_listen_key}. Getting new key...")
                            raise Exception("Listen key refresh failed")

                    # Sleep before next check
                    await self._sleep(self.LISTEN_KEY_RETRY_INTERVAL)
                except asyncio.CancelledError:
                    self.logger().info("Listen key management task cancelled")
                    raise
                except Exception as e:
                    # Reset state on any error to force new key acquisition
                    self.logger().error(f"Error occurred renewing listen key ... {e}")
                    self._current_listen_key = None
                    self._listen_key_initialized_event.clear()
                    await self._sleep(self.LISTEN_KEY_RETRY_INTERVAL)
        finally:
            # Cleanup on task termination
            self.logger().info("Listen key management task stopped")
            await self._ws_assistant.disconnect()
            self._current_listen_key = None
            self._listen_key_initialized_event.clear()

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant

    async def _send_ping(self, websocket_assistant: WSAssistant):
        payload = {
            "method": "PING",
        }
        ping_request: WSJSONRequest = WSJSONRequest(payload=payload)
        await websocket_assistant.send(ping_request)

    async def _on_user_stream_interruption(self, websocket_assistant: Optional[WSAssistant]):
        """
        Handles websocket disconnection by cleaning up resources.

        This method is called when the websocket connection is interrupted.
        It ensures proper cleanup by:
        1. Cancelling the listen key management task
        2. Disconnecting the websocket assistant if it exists
        3. Clearing the current listen key to force renewal
        4. Resetting the initialization event to block new connections

        :param websocket_assistant: The websocket assistant that was disconnected
        """
        self.logger().info("User stream interrupted. Cleaning up...")

        # Cancel listen key management task first
        if self._manage_listen_key_task and not self._manage_listen_key_task.done():
            self._manage_listen_key_task.cancel()
            try:
                await self._manage_listen_key_task
            except asyncio.CancelledError:
                pass
            except Exception:
                pass  # Ignore any exception from the task
            self._manage_listen_key_task = None

        # Disconnect the websocket if it exists
        websocket_assistant and await websocket_assistant.disconnect()
        # Force new listen key acquisition on reconnection
        self._current_listen_key = None
        self._listen_key_initialized_event.clear()

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        while True:
            try:
                await asyncio.wait_for(
                    super()._process_websocket_messages(websocket_assistant=websocket_assistant, queue=queue),
                    timeout=CONSTANTS.WS_CONNECTION_TIME_INTERVAL
                )
            except asyncio.TimeoutError:
                ping_request = WSJSONRequest(payload={"method": "PING"})
                await websocket_assistant.send(ping_request)



import logging
from typing import Any, Dict, List, Optional

from hummingbot.client.config.config_helpers import ClientConfigAdapter, get_connector_class
from hummingbot.client.config.security import Security
from hummingbot.client.settings import AllConnectorSettings
from hummingbot.connector.exchange.paper_trade import create_paper_trade_market
from hummingbot.connector.exchange_base import ExchangeBase


class ConnectorManager:
    """
    Manages connectors (exchanges) dynamically.

    This class provides functionality to:
    - Create and initialize connectors on the fly
    - Add/remove connectors dynamically
    - Access market data without strategies
    - Place orders directly through connectors
    - Manage connector lifecycle independently of strategies
    """

    def __init__(self, client_config: ClientConfigAdapter):
        """
        Initialize the connector manager.

        Args:
            client_config: Client configuration
        """
        self._logger = logging.getLogger(__name__)
        self.client_config_map = client_config

        # Active connectors
        self.connectors: Dict[str, ExchangeBase] = {}

    def create_connector(self,
                         connector_name: str,
                         trading_pairs: List[str],
                         trading_required: bool = True,
                         api_keys: Optional[Dict[str, str]] = None) -> ExchangeBase:
        """
        Create and initialize a connector.

        Args:
            connector_name: Name of the connector (e.g., 'binance', 'kucoin')
            trading_pairs: List of trading pairs to support
            trading_required: Whether this connector will be used for trading
            api_keys: Optional API keys dict

        Returns:
            ExchangeBase: Initialized connector instance
        """
        try:
            # Check if connector already exists
            if connector_name in self.connectors:
                self._logger.warning(f"Connector {connector_name} already exists")
                return self.connectors[connector_name]

            # Handle paper trading connector names
            if connector_name.endswith("_paper_trade"):
                base_connector_name = connector_name.replace("_paper_trade", "")
                conn_setting = AllConnectorSettings.get_connector_settings()[base_connector_name]
            else:
                base_connector_name = connector_name
                conn_setting = AllConnectorSettings.get_connector_settings()[connector_name]

            # Handle paper trading
            if connector_name.endswith("paper_trade"):

                base_connector = base_connector_name
                connector = create_paper_trade_market(
                    base_connector,
                    trading_pairs
                )

                # Set paper trade balances if configured
                paper_trade_account_balance = self.client_config_map.paper_trade.paper_trade_account_balance
                if paper_trade_account_balance is not None:
                    for asset, balance in paper_trade_account_balance.items():
                        connector.set_balance(asset, balance)
            else:
                # Create live connector
                keys = api_keys or Security.api_keys(connector_name)
                if not keys and not conn_setting.uses_gateway_generic_connector():
                    raise ValueError(f"API keys required for live trading connector '{connector_name}'. "
                                     f"Either provide API keys or use a paper trade connector.")

                init_params = conn_setting.conn_init_parameters(
                    trading_pairs=trading_pairs,
                    trading_required=trading_required,
                    api_keys=keys,
                    balance_asset_limit=self.client_config_map.hb_config.balance_asset_limit,
                    rate_limits_share_pct=self.client_config_map.hb_config.rate_limits_share_pct,
                    gateway_config=self.client_config_map.hb_config.gateway,
                )

                connector_class = get_connector_class(connector_name)
                connector = connector_class(**init_params)

            # Add to active connectors
            self.connectors[connector_name] = connector

            self._logger.info(f"Created connector: {connector_name}")

            return connector

        except Exception as e:
            self._logger.error(f"Failed to create connector {connector_name}: {e}")
            raise

    def remove_connector(self, connector_name: str) -> bool:
        """
        Remove a connector and clean up resources.

        Args:
            connector_name: Name of the connector to remove

        Returns:
            bool: True if successfully removed
        """
        if connector_name not in self.connectors:
            self._logger.warning(f"Connector {connector_name} not found")
            return False

        del self.connectors[connector_name]
        self._logger.info(f"Removed connector: {connector_name}")
        return True

    async def add_trading_pairs(self, connector_name: str, trading_pairs: List[str]) -> bool:
        """
        Add trading pairs to an existing connector.

        Args:
            connector_name: Name of the connector
            trading_pairs: List of trading pairs to add

        Returns:
            bool: True if successfully added
        """
        if connector_name not in self.connectors:
            self._logger.error(f"Connector {connector_name} not found")
            return False

        # Most connectors require recreation to add pairs
        # So we'll recreate with the combined list
        connector = self.connectors[connector_name]
        existing_pairs = connector.trading_pairs
        all_pairs = list(set(existing_pairs + trading_pairs))

        # Remove and recreate
        self.remove_connector(connector_name)
        self.create_connector(connector_name, all_pairs)

        return True

    @staticmethod
    def is_gateway_market(connector_name: str) -> bool:
        return connector_name in AllConnectorSettings.get_gateway_amm_connector_names()

    def get_connector(self, connector_name: str) -> Optional[ExchangeBase]:
        """Get a connector by name."""
        return self.connectors.get(connector_name)

    def get_all_connectors(self) -> Dict[str, ExchangeBase]:
        """Get all active connectors."""
        return self.connectors.copy()

    def get_order_book(self, connector_name: str, trading_pair: str) -> Any:
        """Get order book for a trading pair."""
        connector = self.get_connector(connector_name)
        if not connector:
            return None

        return connector.get_order_book(trading_pair)

    def get_balance(self, connector_name: str, asset: str) -> float:
        """Get balance for an asset."""
        connector = self.get_connector(connector_name)
        if not connector:
            return 0.0

        return connector.get_balance(asset)

    def get_all_balances(self, connector_name: str) -> Dict[str, float]:
        """Get all balances from a connector."""
        connector = self.get_connector(connector_name)
        if not connector:
            return {}

        return connector.get_all_balances()

    async def update_connector_balances(self, connector_name: str):
        """
        Update balances for a specific connector.

        Args:
            connector_name: Name of the connector to update balances for
        """
        connector = self.get_connector(connector_name)
        if connector:
            await connector._update_balances()
        else:
            raise ValueError(f"Connector {connector_name} not found")

    def get_status(self) -> Dict[str, Any]:
        """Get status of all connectors."""
        status = {}
        for name, connector in self.connectors.items():
            status[name] = {
                'ready': connector.ready,
                'trading_pairs': connector.trading_pairs,
                'orders_count': len(connector.limit_orders),
                'balances': connector.get_all_balances() if connector.ready else {}
            }
        return status



import decimal
from decimal import Decimal
from typing import Optional

from hummingbot.client.config.config_validators import (
    validate_bool,
    validate_connector,
    validate_decimal,
    validate_exchange,
    validate_int,
    validate_market_trading_pair,
)
from hummingbot.client.config.config_var import ConfigVar
from hummingbot.client.settings import AllConnectorSettings, required_exchanges


def maker_trading_pair_prompt():
    exchange = pure_market_making_config_map.get("exchange").value
    example = AllConnectorSettings.get_example_pairs().get(exchange)
    return "Enter the token trading pair you would like to trade on %s%s >>> " \
           % (exchange, f" (e.g. {example})" if example else "")


# strategy specific validators
def validate_exchange_trading_pair(value: str) -> Optional[str]:
    exchange = pure_market_making_config_map.get("exchange").value
    return validate_market_trading_pair(exchange, value)


def order_amount_prompt() -> str:
    trading_pair = pure_market_making_config_map["market"].value
    base_asset, quote_asset = trading_pair.split("-")
    return f"What is the amount of {base_asset} per order? >>> "


def validate_price_source(value: str) -> Optional[str]:
    if value not in {"current_market", "external_market", "custom_api"}:
        return "Invalid price source type."


def on_validate_price_source(value: str):
    if value != "external_market":
        pure_market_making_config_map["price_source_exchange"].value = None
        pure_market_making_config_map["price_source_market"].value = None
        pure_market_making_config_map["take_if_crossed"].value = None
    if value != "custom_api":
        pure_market_making_config_map["price_source_custom_api"].value = None
    else:
        pure_market_making_config_map["price_type"].value = "custom"


def price_source_market_prompt() -> str:
    external_market = pure_market_making_config_map.get("price_source_exchange").value
    return f'Enter the token trading pair on {external_market} >>> '


def validate_price_source_exchange(value: str) -> Optional[str]:
    if value == pure_market_making_config_map.get("exchange").value:
        return "Price source exchange cannot be the same as maker exchange."
    return validate_connector(value)


def on_validated_price_source_exchange(value: str):
    if value is None:
        pure_market_making_config_map["price_source_market"].value = None


def validate_price_source_market(value: str) -> Optional[str]:
    market = pure_market_making_config_map.get("price_source_exchange").value
    return validate_market_trading_pair(market, value)


def validate_price_floor_ceiling(value: str) -> Optional[str]:
    try:
        decimal_value = Decimal(value)
    except Exception:
        return f"{value} is not in decimal format."
    if not (decimal_value == Decimal("-1") or decimal_value > Decimal("0")):
        return "Value must be more than 0 or -1 to disable this feature."


def validate_price_type(value: str) -> Optional[str]:
    error = None
    price_source = pure_market_making_config_map.get("price_source").value
    if price_source != "custom_api":
        valid_values = {"mid_price",
                        "last_price",
                        "last_own_trade_price",
                        "best_bid",
                        "best_ask",
                        "inventory_cost",
                        }
        if value not in valid_values:
            error = "Invalid price type."
    elif value != "custom":
        error = "Invalid price type."
    return error


def on_validated_price_type(value: str):
    if value == 'inventory_cost':
        pure_market_making_config_map["inventory_price"].value = None


def exchange_on_validated(value: str):
    required_exchanges.add(value)


def validate_decimal_list(value: str) -> Optional[str]:
    decimal_list = list(value.split(","))
    for number in decimal_list:
        try:
            validate_result = validate_decimal(Decimal(number), 0, 100, inclusive=False)
        except decimal.InvalidOperation:
            return "Please enter valid decimal numbers"
        if validate_result is not None:
            return validate_result


pure_market_making_config_map = {
    "strategy":
        ConfigVar(key="strategy",
                  prompt=None,
                  default="pure_market_making"),
    "exchange":
        ConfigVar(key="exchange",
                  prompt="Enter your maker spot connector >>> ",
                  validator=validate_exchange,
                  on_validated=exchange_on_validated,
                  prompt_on_new=True),
    "market":
        ConfigVar(key="market",
                  prompt=maker_trading_pair_prompt,
                  validator=validate_exchange_trading_pair,
                  prompt_on_new=True),
    "bid_spread":
        ConfigVar(key="bid_spread",
                  prompt="How far away from the mid price do you want to place the "
                         "first bid order? (Enter 1 to indicate 1%) >>> ",
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, 0, 100, inclusive=False),
                  prompt_on_new=True),
    "ask_spread":
        ConfigVar(key="ask_spread",
                  prompt="How far away from the mid price do you want to place the "
                         "first ask order? (Enter 1 to indicate 1%) >>> ",
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, 0, 100, inclusive=False),
                  prompt_on_new=True),
    "minimum_spread":
        ConfigVar(key="minimum_spread",
                  prompt="At what minimum spread should the bot automatically cancel orders? (Enter 1 for 1%) >>> ",
                  required_if=lambda: False,
                  type_str="decimal",
                  default=Decimal(-100),
                  validator=lambda v: validate_decimal(v, -100, 100, True)),
    "order_refresh_time":
        ConfigVar(key="order_refresh_time",
                  prompt="How often do you want to cancel and replace bids and asks "
                         "(in seconds)? >>> ",
                  type_str="float",
                  validator=lambda v: validate_decimal(v, 0, inclusive=False),
                  prompt_on_new=True),
    "max_order_age":
        ConfigVar(key="max_order_age",
                  prompt="How long do you want to cancel and replace bids and asks "
                         "with the same price (in seconds)? >>> ",
                  type_str="float",
                  default=Decimal("1800"),
                  validator=lambda v: validate_decimal(v, 0, inclusive=False)),
    "order_refresh_tolerance_pct":
        ConfigVar(key="order_refresh_tolerance_pct",
                  prompt="Enter the percent change in price needed to refresh orders at each cycle "
                         "(Enter 1 to indicate 1%) >>> ",
                  type_str="decimal",
                  default=Decimal("0"),
                  validator=lambda v: validate_decimal(v, -10, 10, inclusive=True)),
    "order_amount":
        ConfigVar(key="order_amount",
                  prompt=order_amount_prompt,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, min_value=Decimal("0"), inclusive=False),
                  prompt_on_new=True),
    "price_ceiling":
        ConfigVar(key="price_ceiling",
                  prompt="Enter the price point above which only sell orders will be placed "
                         "(Enter -1 to deactivate this feature) >>> ",
                  type_str="decimal",
                  default=Decimal("-1"),
                  validator=validate_price_floor_ceiling),
    "price_floor":
        ConfigVar(key="price_floor",
                  prompt="Enter the price below which only buy orders will be placed "
                         "(Enter -1 to deactivate this feature) >>> ",
                  type_str="decimal",
                  default=Decimal("-1"),
                  validator=validate_price_floor_ceiling),
    "moving_price_band_enabled":
        ConfigVar(key="moving_price_band_enabled",
                  prompt="Would you like to enable moving price floor and ceiling? (Yes/No) >>> ",
                  type_str="bool",
                  default=False,
                  validator=validate_bool),
    "price_ceiling_pct":
        ConfigVar(key="price_ceiling_pct",
                  prompt="Enter a percentage to the current price that sets the price ceiling. Above this price, only sell orders will be placed >>> ",
                  type_str="decimal",
                  default=Decimal("1"),
                  required_if=lambda: pure_market_making_config_map.get("moving_price_band_enabled").value,
                  validator=validate_decimal),
    "price_floor_pct":
        ConfigVar(key="price_floor_pct",
                  prompt="Enter a percentage to the current price that sets the price floor. Below this price, only buy orders will be placed >>> ",
                  type_str="decimal",
                  default=Decimal("-1"),
                  required_if=lambda: pure_market_making_config_map.get("moving_price_band_enabled").value,
                  validator=validate_decimal),
    "price_band_refresh_time":
        ConfigVar(key="price_band_refresh_time",
                  prompt="After this amount of time (in seconds), the price bands are reset based on the current price >>> ",
                  type_str="float",
                  default=86400,
                  required_if=lambda: pure_market_making_config_map.get("moving_price_band_enabled").value,
                  validator=validate_decimal),
    "ping_pong_enabled":
        ConfigVar(key="ping_pong_enabled",
                  prompt="Would you like to use the ping pong feature and alternate between buy and sell orders after fills? (Yes/No) >>> ",
                  type_str="bool",
                  default=False,
                  prompt_on_new=True,
                  validator=validate_bool),
    "order_levels":
        ConfigVar(key="order_levels",
                  prompt="How many orders do you want to place on both sides? >>> ",
                  type_str="int",
                  validator=lambda v: validate_int(v, min_value=-1, inclusive=False),
                  default=1),
    "order_level_amount":
        ConfigVar(key="order_level_amount",
                  prompt="How much do you want to increase or decrease the order size for each "
                         "additional order? (decrease < 0 > increase) >>> ",
                  required_if=lambda: pure_market_making_config_map.get("order_levels").value > 1,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v),
                  default=0),
    "order_level_spread":
        ConfigVar(key="order_level_spread",
                  prompt="Enter the price increments (as percentage) for subsequent "
                         "orders? (Enter 1 to indicate 1%) >>> ",
                  required_if=lambda: pure_market_making_config_map.get("order_levels").value > 1,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, 0, 100, inclusive=False),
                  default=Decimal("1")),
    "inventory_skew_enabled":
        ConfigVar(key="inventory_skew_enabled",
                  prompt="Would you like to enable inventory skew? (Yes/No) >>> ",
                  type_str="bool",
                  default=False,
                  validator=validate_bool),
    "inventory_target_base_pct":
        ConfigVar(key="inventory_target_base_pct",
                  prompt="What is your target base asset percentage? Enter 50 for 50% >>> ",
                  required_if=lambda: pure_market_making_config_map.get("inventory_skew_enabled").value,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, 0, 100),
                  default=Decimal("50")),
    "inventory_range_multiplier":
        ConfigVar(key="inventory_range_multiplier",
                  prompt="What is your tolerable range of inventory around the target, "
                         "expressed in multiples of your total order size? ",
                  required_if=lambda: pure_market_making_config_map.get("inventory_skew_enabled").value,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, min_value=0, inclusive=False),
                  default=Decimal("1")),
    "inventory_price":
        ConfigVar(key="inventory_price",
                  prompt="What is the price of your base asset inventory? ",
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, min_value=Decimal("0"), inclusive=True),
                  required_if=lambda: pure_market_making_config_map.get("price_type").value == "inventory_cost",
                  default=Decimal("1"),
                  ),
    "filled_order_delay":
        ConfigVar(key="filled_order_delay",
                  prompt="How long do you want to wait before placing the next order "
                         "if your order gets filled (in seconds)? >>> ",
                  type_str="float",
                  validator=lambda v: validate_decimal(v, min_value=0, inclusive=False),
                  default=60),
    "hanging_orders_enabled":
        ConfigVar(key="hanging_orders_enabled",
                  prompt="Do you want to enable hanging orders? (Yes/No) >>> ",
                  type_str="bool",
                  default=False,
                  validator=validate_bool),
    "hanging_orders_cancel_pct":
        ConfigVar(key="hanging_orders_cancel_pct",
                  prompt="At what spread percentage (from mid price) will hanging orders be canceled? "
                         "(Enter 1 to indicate 1%) >>> ",
                  required_if=lambda: pure_market_making_config_map.get("hanging_orders_enabled").value,
                  type_str="decimal",
                  default=Decimal("10"),
                  validator=lambda v: validate_decimal(v, 0, 100, inclusive=False)),
    "order_optimization_enabled":
        ConfigVar(key="order_optimization_enabled",
                  prompt="Do you want to enable best bid ask jumping? (Yes/No) >>> ",
                  type_str="bool",
                  default=False,
                  validator=validate_bool),
    "ask_order_optimization_depth":
        ConfigVar(key="ask_order_optimization_depth",
                  prompt="How deep do you want to go into the order book for calculating "
                         "the top ask, ignoring dust orders on the top "
                         "(expressed in base asset amount)? >>> ",
                  required_if=lambda: pure_market_making_config_map.get("order_optimization_enabled").value,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, min_value=0),
                  default=0),
    "bid_order_optimization_depth":
        ConfigVar(key="bid_order_optimization_depth",
                  prompt="How deep do you want to go into the order book for calculating "
                         "the top bid, ignoring dust orders on the top "
                         "(expressed in base asset amount)? >>> ",
                  required_if=lambda: pure_market_making_config_map.get("order_optimization_enabled").value,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, min_value=0),
                  default=0),
    "add_transaction_costs":
        ConfigVar(key="add_transaction_costs",
                  prompt="Do you want to add transaction costs automatically to order prices? (Yes/No) >>> ",
                  type_str="bool",
                  default=False,
                  validator=validate_bool),
    "price_source":
        ConfigVar(key="price_source",
                  prompt="Which price source to use? (current_market/external_market/custom_api) >>> ",
                  type_str="str",
                  default="current_market",
                  validator=validate_price_source,
                  on_validated=on_validate_price_source),
    "price_type":
        ConfigVar(key="price_type",
                  prompt="Which price type to use? ("
                         "mid_price/last_price/last_own_trade_price/best_bid/best_ask/inventory_cost) >>> ",
                  type_str="str",
                  required_if=lambda: pure_market_making_config_map.get("price_source").value != "custom_api",
                  default="mid_price",
                  on_validated=on_validated_price_type,
                  validator=validate_price_type),
    "price_source_exchange":
        ConfigVar(key="price_source_exchange",
                  prompt="Enter external price source exchange name >>> ",
                  required_if=lambda: pure_market_making_config_map.get("price_source").value == "external_market",
                  type_str="str",
                  validator=validate_price_source_exchange,
                  on_validated=on_validated_price_source_exchange),
    "price_source_market":
        ConfigVar(key="price_source_market",
                  prompt=price_source_market_prompt,
                  required_if=lambda: pure_market_making_config_map.get("price_source").value == "external_market",
                  type_str="str",
                  validator=validate_price_source_market),
    "take_if_crossed":
        ConfigVar(key="take_if_crossed",
                  prompt="Do you want to take the best order if orders cross the orderbook? ((Yes/No) >>> ",
                  required_if=lambda: pure_market_making_config_map.get(
                      "price_source").value == "external_market",
                  type_str="bool",
                  validator=validate_bool),
    "price_source_custom_api":
        ConfigVar(key="price_source_custom_api",
                  prompt="Enter pricing API URL >>> ",
                  required_if=lambda: pure_market_making_config_map.get("price_source").value == "custom_api",
                  type_str="str"),
    "custom_api_update_interval":
        ConfigVar(key="custom_api_update_interval",
                  prompt="Enter custom API update interval in second (default: 5.0, min: 0.5) >>> ",
                  required_if=lambda: False,
                  default=float(5),
                  type_str="float",
                  validator=lambda v: validate_decimal(v, Decimal("0.5"))),
    "order_override":
        ConfigVar(key="order_override",
                  prompt=None,
                  required_if=lambda: False,
                  default=None,
                  type_str="json"),
    "should_wait_order_cancel_confirmation":
        ConfigVar(key="should_wait_order_cancel_confirmation",
                  prompt="Should the strategy wait to receive a confirmation for orders cancelation "
                         "before creating a new set of orders? "
                         "(Not waiting requires enough available balance) (Yes/No) >>> ",
                  type_str="bool",
                  default=True,
                  validator=validate_bool),
    "split_order_levels_enabled":
        ConfigVar(key="split_order_levels_enabled",
                  prompt="Do you want bid and ask orders to be placed at multiple defined spread and amount? "
                         "This acts as an overrides which replaces order_amount, order_spreads, "
                         "order_level_amount, order_level_spreads (Yes/No) >>> ",
                  default=False,
                  type_str="bool",
                  validator=validate_bool),
    "bid_order_level_spreads":
        ConfigVar(key="bid_order_level_spreads",
                  prompt="Enter the spreads (as percentage) for all bid spreads "
                         "e.g 1,2,3,4 to represent 1%,2%,3%,4%. "
                         "The number of levels set will be equal to the "
                         "minimum length of bid_order_level_spreads and bid_order_level_amounts >>> ",
                  default=None,
                  type_str="str",
                  required_if=lambda: pure_market_making_config_map.get(
                      "split_order_levels_enabled").value,
                  validator=validate_decimal_list),
    "ask_order_level_spreads":
        ConfigVar(key="ask_order_level_spreads",
                  prompt="Enter the spreads (as percentage) for all ask spreads "
                         "e.g 1,2,3,4 to represent 1%,2%,3%,4%. "
                         "The number of levels set will be equal to the "
                         "minimum length of bid_order_level_spreads and bid_order_level_amounts >>> ",
                  default=None,
                  type_str="str",
                  required_if=lambda: pure_market_making_config_map.get(
                      "split_order_levels_enabled").value,
                  validator=validate_decimal_list),
    "bid_order_level_amounts":
        ConfigVar(key="bid_order_level_amounts",
                  prompt="Enter the amount for all bid amounts. "
                         "e.g 1,2,3,4. "
                         "The number of levels set will be equal to the "
                         "minimum length of bid_order_level_spreads and bid_order_level_amounts >>> ",
                  default=None,
                  type_str="str",
                  required_if=lambda: pure_market_making_config_map.get(
                      "split_order_levels_enabled").value,
                  validator=validate_decimal_list),
    "ask_order_level_amounts":
        ConfigVar(key="ask_order_level_amounts",
                  prompt="Enter the amount for all ask amounts. "
                         "e.g 1,2,3,4. "
                         "The number of levels set will be equal to the "
                         "minimum length of bid_order_level_spreads and bid_order_level_amounts >>> ",
                  default=None,
                  required_if=lambda: pure_market_making_config_map.get(
                      "split_order_levels_enabled").value,
                  type_str="str",
                  validator=validate_decimal_list),
}



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class PublicMiniTickerV3Api(_message.Message):
    __slots__ = ("symbol", "price", "rate", "zonedRate", "high", "low", "volume", "quantity", "lastCloseRate", "lastCloseZonedRate", "lastCloseHigh", "lastCloseLow")
    SYMBOL_FIELD_NUMBER: _ClassVar[int]
    PRICE_FIELD_NUMBER: _ClassVar[int]
    RATE_FIELD_NUMBER: _ClassVar[int]
    ZONEDRATE_FIELD_NUMBER: _ClassVar[int]
    HIGH_FIELD_NUMBER: _ClassVar[int]
    LOW_FIELD_NUMBER: _ClassVar[int]
    VOLUME_FIELD_NUMBER: _ClassVar[int]
    QUANTITY_FIELD_NUMBER: _ClassVar[int]
    LASTCLOSERATE_FIELD_NUMBER: _ClassVar[int]
    LASTCLOSEZONEDRATE_FIELD_NUMBER: _ClassVar[int]
    LASTCLOSEHIGH_FIELD_NUMBER: _ClassVar[int]
    LASTCLOSELOW_FIELD_NUMBER: _ClassVar[int]
    symbol: str
    price: str
    rate: str
    zonedRate: str
    high: str
    low: str
    volume: str
    quantity: str
    lastCloseRate: str
    lastCloseZonedRate: str
    lastCloseHigh: str
    lastCloseLow: str
    def __init__(self, symbol: _Optional[str] = ..., price: _Optional[str] = ..., rate: _Optional[str] = ..., zonedRate: _Optional[str] = ..., high: _Optional[str] = ..., low: _Optional[str] = ..., volume: _Optional[str] = ..., quantity: _Optional[str] = ..., lastCloseRate: _Optional[str] = ..., lastCloseZonedRate: _Optional[str] = ..., lastCloseHigh: _Optional[str] = ..., lastCloseLow: _Optional[str] = ...) -> None: ...



from abc import ABC, abstractmethod
from decimal import ROUND_DOWN, Decimal
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

from xrpl.models import XRP, IssuedCurrencyAmount, Memo, OfferCreate, Path, PathStep, Payment, PaymentFlag, Transaction
from xrpl.utils import xrp_to_drops

from hummingbot.connector.exchange.xrpl import xrpl_constants as CONSTANTS
from hummingbot.connector.exchange.xrpl.xrpl_utils import convert_string_to_hex
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder

if TYPE_CHECKING:
    from hummingbot.connector.exchange.xrpl.xrpl_exchange import XrplExchange


class XRPLOrderPlacementStrategy(ABC):
    """Abstract base class for XRPL order placement strategies"""

    def __init__(self, connector: "XrplExchange", order: InFlightOrder):
        self._connector = connector
        self._order = order

    @abstractmethod
    async def create_order_transaction(self) -> Transaction:
        """Create the appropriate transaction for the order type"""
        pass

    def get_base_quote_amounts(
        self, price: Optional[Decimal] = None
    ) -> Tuple[Union[str, IssuedCurrencyAmount], Union[str, IssuedCurrencyAmount]]:
        """Calculate the base and quote amounts for the order"""
        base_currency, quote_currency = self._connector.get_currencies_from_trading_pair(self._order.trading_pair)
        trading_rule = self._connector._trading_rules[self._order.trading_pair]

        amount_in_base_quantum = Decimal(trading_rule.min_base_amount_increment)
        amount_in_quote_quantum = Decimal(trading_rule.min_quote_amount_increment)

        # Use price if provided, otherwise use order price
        effective_price = price if price is not None else self._order.price
        if effective_price is None:
            raise ValueError("Price must be provided either in the order or as a parameter")

        amount_in_base = Decimal(self._order.amount.quantize(amount_in_base_quantum, rounding=ROUND_DOWN))
        amount_in_quote = Decimal(
            (self._order.amount * effective_price).quantize(amount_in_quote_quantum, rounding=ROUND_DOWN)
        )

        # Handle precision for base and quote amounts
        total_digits_base = len(str(amount_in_base).split(".")[1]) + len(str(amount_in_base).split(".")[0])
        if total_digits_base > CONSTANTS.XRPL_MAX_DIGIT:  # XRPL_MAX_DIGIT
            adjusted_quantum = CONSTANTS.XRPL_MAX_DIGIT - len(str(amount_in_base).split(".")[0])
            amount_in_base = Decimal(amount_in_base.quantize(Decimal(f"1e-{adjusted_quantum}"), rounding=ROUND_DOWN))

        total_digits_quote = len(str(amount_in_quote).split(".")[1]) + len(str(amount_in_quote).split(".")[0])
        if total_digits_quote > CONSTANTS.XRPL_MAX_DIGIT:  # XRPL_MAX_DIGIT
            adjusted_quantum = CONSTANTS.XRPL_MAX_DIGIT - len(str(amount_in_quote).split(".")[0])
            amount_in_quote = Decimal(amount_in_quote.quantize(Decimal(f"1e-{adjusted_quantum}"), rounding=ROUND_DOWN))

        # Convert amounts based on currency type
        if self._order.trade_type is TradeType.SELL:
            if isinstance(base_currency, XRP):
                we_pay = xrp_to_drops(amount_in_base)
            else:
                we_pay = IssuedCurrencyAmount(
                    currency=base_currency.currency, issuer=base_currency.issuer, value=str(amount_in_base)
                )

            if isinstance(quote_currency, XRP):
                we_get = xrp_to_drops(amount_in_quote)
            else:
                we_get = IssuedCurrencyAmount(
                    currency=quote_currency.currency, issuer=quote_currency.issuer, value=str(amount_in_quote)
                )
        else:
            if isinstance(quote_currency, XRP):
                we_pay = xrp_to_drops(amount_in_quote)
            else:
                we_pay = IssuedCurrencyAmount(
                    currency=quote_currency.currency, issuer=quote_currency.issuer, value=str(amount_in_quote)
                )

            if isinstance(base_currency, XRP):
                we_get = xrp_to_drops(amount_in_base)
            else:
                we_get = IssuedCurrencyAmount(
                    currency=base_currency.currency, issuer=base_currency.issuer, value=str(amount_in_base)
                )

        return we_pay, we_get


class LimitOrderStrategy(XRPLOrderPlacementStrategy):
    """Strategy for placing limit orders"""

    async def create_order_transaction(self) -> Transaction:
        we_pay, we_get = self.get_base_quote_amounts()
        flags = self._connector.xrpl_order_type(self._order.order_type)

        flags += CONSTANTS.XRPL_SELL_FLAG

        memo = Memo(memo_data=convert_string_to_hex(self._order.client_order_id, padding=False))
        return OfferCreate(
            account=self._connector._xrpl_auth.get_account(),
            flags=flags,
            taker_gets=we_pay,
            taker_pays=we_get,
            memos=[memo],
        )


class MarketOrderStrategy(XRPLOrderPlacementStrategy):
    """Strategy for placing market orders"""

    async def create_order_transaction(self) -> Transaction:
        # Get best price from order book
        price = Decimal(
            await self._connector._get_best_price(
                self._order.trading_pair, is_buy=True if self._order.trade_type is TradeType.BUY else False
            )
        )

        # Add slippage to make sure we get the order filled
        if self._order.trade_type is TradeType.SELL:
            price *= Decimal("1") - CONSTANTS.MARKET_ORDER_MAX_SLIPPAGE
        else:
            price *= Decimal("1") + CONSTANTS.MARKET_ORDER_MAX_SLIPPAGE

        we_pay, we_get = self.get_base_quote_amounts(price)
        flags = self._connector.xrpl_order_type(self._order.order_type)

        flags += CONSTANTS.XRPL_SELL_FLAG

        memo = Memo(memo_data=convert_string_to_hex(self._order.client_order_id, padding=False))
        return OfferCreate(
            account=self._connector._xrpl_auth.get_account(),
            flags=flags,
            taker_gets=we_pay,
            taker_pays=we_get,
            memos=[memo],
        )


class AMMSwapOrderStrategy(XRPLOrderPlacementStrategy):
    """Strategy for placing AMM swap orders"""

    async def create_order_transaction(self) -> Transaction:
        # Get best price from order book
        price = Decimal(
            await self._connector._get_best_price(
                self._order.trading_pair, is_buy=True if self._order.trade_type is TradeType.BUY else False
            )
        )

        fee_rate_pct = self._connector._trading_pair_fee_rules[self._order.trading_pair].get(
            "amm_pool_fee", Decimal("0.0")
        )

        we_pay, we_get = self.get_base_quote_amounts(price)

        if self._order.trade_type is TradeType.BUY:
            # add slippage to we_get
            if isinstance(we_get, IssuedCurrencyAmount):
                we_get = IssuedCurrencyAmount(
                    currency=we_get.currency,
                    issuer=we_get.issuer,
                    value=str(Decimal(we_get.value) * Decimal("1") + fee_rate_pct),
                )
            else:
                we_get = str(int(Decimal(we_get) * Decimal("1") + fee_rate_pct))

            if isinstance(we_pay, IssuedCurrencyAmount):
                we_pay = IssuedCurrencyAmount(
                    currency=we_pay.currency,
                    issuer=we_pay.issuer,
                    value=str(Decimal(we_pay.value) * Decimal("1") + fee_rate_pct),
                )
            else:
                we_pay = str(int(Decimal(we_pay) * Decimal("1") + fee_rate_pct))
        else:
            we_pay, we_get = self.get_base_quote_amounts(price * Decimal(1 + fee_rate_pct))

        paths: Optional[List[Path]] = None

        # if both we_pay and we_get are not XRP:
        if isinstance(we_pay, IssuedCurrencyAmount) and isinstance(we_get, IssuedCurrencyAmount):
            path: Path = [
                PathStep(
                    account=we_pay.issuer,
                ),
                PathStep(
                    currency=we_get.currency,
                    issuer=we_get.issuer,
                ),
            ]
            paths = [path]

        # if we_pay is XRP, we_get must be an IssuedCurrencyAmount
        if isinstance(we_pay, str) and isinstance(we_get, IssuedCurrencyAmount):
            path: Path = [
                PathStep(
                    currency=we_get.currency,
                    issuer=we_get.issuer,
                ),
            ]
            paths = [path]

        # if we_pay is IssuedCurrencyAmount, we_get must be XRP
        if isinstance(we_pay, IssuedCurrencyAmount) and isinstance(we_get, str):
            path: Path = [
                PathStep(currency="XRP"),
            ]
            paths = [path]

        swap_amm_prefix = "AMM_SWAP"

        memo = Memo(memo_data=convert_string_to_hex(f"{self._order.client_order_id}_{swap_amm_prefix}", padding=False))

        return Payment(
            account=self._connector._xrpl_auth.get_account(),
            destination=self._connector._xrpl_auth.get_account(),
            amount=we_get,
            send_max=we_pay,
            paths=paths,
            memos=[memo],
            flags=PaymentFlag.TF_NO_RIPPLE_DIRECT + PaymentFlag.TF_PARTIAL_PAYMENT,
        )


class OrderPlacementStrategyFactory:
    """Factory for creating order placement strategies"""

    @staticmethod
    def create_strategy(connector: "XrplExchange", order: InFlightOrder) -> XRPLOrderPlacementStrategy:
        if order.order_type == OrderType.LIMIT or order.order_type == OrderType.LIMIT_MAKER:
            return LimitOrderStrategy(connector, order)
        elif order.order_type == OrderType.MARKET:
            return MarketOrderStrategy(connector, order)
        elif order.order_type == OrderType.AMM_SWAP:
            return AMMSwapOrderStrategy(connector, order)
        else:
            raise ValueError(f"Unsupported order type: {order.order_type}")



import asyncio
import threading
from decimal import Decimal
from typing import TYPE_CHECKING, Dict, List

import pandas as pd

from hummingbot.client.config.config_validators import validate_decimal, validate_exchange
from hummingbot.client.performance import PerformanceMetrics
from hummingbot.client.settings import AllConnectorSettings
from hummingbot.core.rate_oracle.rate_oracle import RateOracle
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.user.user_balances import UserBalances

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401

OPTIONS = [
    "limit",
    "paper"
]


class BalanceCommand:
    def balance(self,  # type: HummingbotApplication
                option: str = None,
                args: List[str] = None
                ):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.balance, option, args)
            return

        self.app.clear_input()
        if option is None:
            safe_ensure_future(self.show_balances())

        elif option in OPTIONS:
            if option == "limit":
                balance_asset_limit = self.client_config_map.balance_asset_limit
                if args is None or len(args) == 0:
                    safe_ensure_future(self.show_asset_limits())
                    return
                if len(args) != 3 or validate_exchange(args[0]) is not None or validate_decimal(args[2]) is not None:
                    self.notify("Error: Invalid command arguments")
                    self.notify_balance_limit_set()
                    return
                exchange = args[0]
                asset = args[1].upper()
                amount = float(args[2])
                if balance_asset_limit.get(exchange) is None:
                    balance_asset_limit[exchange] = {}
                if amount < 0 and asset in balance_asset_limit[exchange].keys():
                    balance_asset_limit[exchange].pop(asset)
                    self.notify(f"Limit for {asset} on {exchange} exchange removed.")
                elif amount >= 0:
                    balance_asset_limit[exchange][asset] = amount
                    self.notify(f"Limit for {asset} on {exchange} exchange set to {amount}")
                self.save_client_config()

            elif option == "paper":
                paper_balances = self.client_config_map.paper_trade.paper_trade_account_balance
                if args is None or len(args) == 0:
                    safe_ensure_future(self.show_paper_account_balance())
                    return
                if len(args) != 2 or validate_decimal(args[1]) is not None:
                    self.notify("Error: Invalid command arguments")
                    self.notify_balance_paper_set()
                    return
                asset = args[0].upper()
                amount = float(args[1])
                paper_balances[asset] = amount
                self.notify(f"Paper balance for {asset} token set to {amount}")
                self.save_client_config()

    async def show_balances(
        self  # type: HummingbotApplication
    ):
        global_token_symbol = self.client_config_map.global_token.global_token_symbol
        total_col_name = f"Total ({global_token_symbol})"
        sum_not_for_show_name = "sum_not_for_show"
        self.notify("Updating balances, please wait...")
        network_timeout = float(self.client_config_map.commands_timeout.other_commands_timeout)
        try:
            all_ex_bals = await asyncio.wait_for(
                UserBalances.instance().all_balances_all_exchanges(self.client_config_map), network_timeout
            )
        except asyncio.TimeoutError:
            self.notify("\nA network error prevented the balances to update. See logs for more details.")
            raise
        all_ex_avai_bals = UserBalances.instance().all_available_balances_all_exchanges()

        exchanges_total = 0

        for exchange, bals in all_ex_bals.items():
            self.notify(f"\n{exchange}:")
            df, allocated_total = await self.exchange_balances_extra_df(exchange, bals, all_ex_avai_bals.get(exchange, {}))
            if df.empty:
                self.notify("You have no balance on this exchange.")
            else:
                lines = [
                    "    " + line for line in df.drop(sum_not_for_show_name, axis=1).to_string(index=False).split("\n")
                ]
                self.notify("\n".join(lines))
                self.notify(f"\n  Total: {global_token_symbol} "
                            f"{PerformanceMetrics.smart_round(df[total_col_name].sum())}")
                allocated_percentage = 0
                if df[sum_not_for_show_name].sum() != Decimal("0"):
                    allocated_percentage = allocated_total / df[sum_not_for_show_name].sum()
                self.notify(f"Allocated: {allocated_percentage:.2%}")
                exchanges_total += df[total_col_name].sum()

        self.notify(f"\n\nExchanges Total: {global_token_symbol} {exchanges_total:.0f}    ")

    async def exchange_balances_extra_df(self,  # type: HummingbotApplication
                                         exchange: str,
                                         ex_balances: Dict[str, Decimal],
                                         ex_avai_balances: Dict[str, Decimal]):
        conn_setting = AllConnectorSettings.get_connector_settings()[exchange]
        global_token_symbol = self.client_config_map.global_token.global_token_symbol
        total_col_name = f"Total ({global_token_symbol})"
        allocated_total = Decimal("0")
        rows = []
        for token, bal in ex_balances.items():
            avai = Decimal(ex_avai_balances.get(token.upper(), 0)) if ex_avai_balances is not None else Decimal(0)
            # show zero balances if it is a gateway connector (the user manually
            # chose to show those values with 'gateway connector-tokens')
            if conn_setting.uses_gateway_generic_connector():
                if bal == Decimal(0):
                    allocated = "0%"
                else:
                    allocated = f"{(bal - avai) / bal:.0%}"
            else:
                # the exchange is CEX. Only show balance if non-zero.
                if bal == Decimal(0):
                    continue
                allocated = f"{(bal - avai) / bal:.0%}"

            rate = await RateOracle.get_instance().get_rate(base_token=token)
            rate = Decimal("0") if rate is None else rate
            global_value = rate * bal
            allocated_total += rate * (bal - avai)
            rows.append({"Asset": token.upper(),
                         "Total": round(bal, 4),
                         total_col_name: PerformanceMetrics.smart_round(global_value),
                         "sum_not_for_show": global_value,
                         "Allocated": allocated,
                         })
        df = pd.DataFrame(data=rows, columns=["Asset", "Total", total_col_name, "sum_not_for_show", "Allocated"])
        df.sort_values(by=["Asset"], inplace=True)
        return df, allocated_total

    async def asset_limits_df(self,
                              asset_limit_conf: Dict[str, str]):
        rows = []
        for token, amount in asset_limit_conf.items():
            rows.append({"Asset": token, "Limit": round(Decimal(amount), 4)})

        df = pd.DataFrame(data=rows, columns=["Asset", "Limit"])
        df.sort_values(by=["Asset"], inplace=True)
        return df

    async def show_asset_limits(
        self  # type: HummingbotApplication
    ):
        exchange_limit_conf = self.client_config_map.balance_asset_limit

        if not any(list(exchange_limit_conf.values())):
            self.notify("You have not set any limits.")
            self.notify_balance_limit_set()
            return

        self.notify("Balance Limits per exchange...")

        for exchange, asset_limit_config in exchange_limit_conf.items():
            if asset_limit_config is None:
                continue

            self.notify(f"\n{exchange}")
            df = await self.asset_limits_df(asset_limit_config)
            if df.empty:
                self.notify("You have no limits on this exchange.")
            else:
                lines = ["    " + line for line in df.to_string(index=False).split("\n")]
                self.notify("\n".join(lines))
        self.notify("\n")
        return

    async def paper_acccount_balance_df(self, paper_balances: Dict[str, Decimal]):
        rows = []
        for asset, balance in paper_balances.items():
            rows.append({"Asset": asset, "Balance": round(Decimal(str(balance)), 4)})
        df = pd.DataFrame(data=rows, columns=["Asset", "Balance"])
        df.sort_values(by=["Asset"], inplace=True)
        return df

    def notify_balance_limit_set(self):
        self.notify("To set a balance limit (how much the bot can use): \n"
                    "    balance limit [EXCHANGE] [ASSET] [AMOUNT]\n"
                    "e.g. balance limit binance BTC 0.1")

    def notify_balance_paper_set(self):
        self.notify("To set a paper account balance: \n"
                    "    balance paper [ASSET] [AMOUNT]\n"
                    "e.g. balance paper BTC 0.1")

    async def show_paper_account_balance(
        self  # type: HummingbotApplication
    ):
        paper_balances = self.client_config_map.paper_trade.paper_trade_account_balance
        if not paper_balances:
            self.notify("You have not set any paper account balance.")
            self.notify_balance_paper_set()
            return
        self.notify("Paper account balances:")
        df = await self.paper_acccount_balance_df(paper_balances)
        lines = ["    " + line for line in df.to_string(index=False).split("\n")]
        self.notify("\n".join(lines))
        self.notify("\n")
        return



import numpy as np
from libc.stdint cimport int64_t
cimport numpy as np

cdef class RingBuffer:
    cdef:
        np.float64_t[:] _buffer
        int64_t _delimiter
        int64_t _length
        bint _is_full

    cdef void c_add_value(self, float val)
    cdef void c_increment_delimiter(self)
    cdef double c_get_last_value(self)
    cdef bint c_is_full(self)
    cdef bint c_is_empty(self)
    cdef double c_mean_value(self)
    cdef double c_variance(self)
    cdef double c_std_dev(self)
    cdef np.ndarray[np.double_t, ndim=1] c_get_as_numpy_array(self)



import asyncio
import logging
import time
from typing import Optional

import hummingbot.connector.exchange.bing_x.bing_x_constants as CONSTANTS
import hummingbot.connector.exchange.bing_x.bing_x_utils as utils
import hummingbot.connector.exchange.bing_x.bing_x_web_utils as web_utils
from hummingbot.connector.exchange.bing_x.bing_x_auth import BingXAuth
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger


class BingXAPIUserStreamDataSource(UserStreamTrackerDataSource):

    LISTEN_KEY_KEEP_ALIVE_INTERVAL = 1800

    _bausds_logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 auth: BingXAuth,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN,
                 api_factory: Optional[WebAssistantsFactory] = None,
                 throttler: Optional[AsyncThrottler] = None):
        super().__init__()
        self._auth: BingXAuth = auth
        self._last_recv_time: float = 0
        self._domain = domain
        self._throttler = throttler
        self._api_factory = api_factory or web_utils.build_api_factory(
            throttler=self._throttler,
            domain=self._domain,
            auth=self._auth)
        self._ws_assistant: Optional[WSAssistant] = None
        self._last_ws_message_sent_timestamp = 0

        self._listen_key_initialized_event: asyncio.Event = asyncio.Event()
        self._last_listen_key_ping_ts = 0
        self._current_listen_key = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._bausds_logger is None:
            cls._bausds_logger = logging.getLogger(__name__)
        return cls._bausds_logger

    @property
    def last_recv_time(self) -> float:
        """
        Returns the time of the last received message
        :return: the timestamp of the last received message in seconds
        """
        if self._ws_assistant:
            return self._ws_assistant.last_recv_time
        return 0

    async def listen_for_user_stream(self, output: asyncio.Queue):
        """
        Connects to the user private channel in the exchange using a websocket connection. With the established
        connection listens to all balance events and order updates provided by the exchange, and stores them in the
        output queue
        :param output: the queue to use to store the received messages
        """
        ws = None
        while True:
            try:
                ws: WSAssistant = await self._connected_websocket_assistant()
                # await ws.connect(ws_url=CONSTANTS.WSS_PRIVATE_URL[self._domain])
                # await self._authenticate_connection(ws)
                await self._subscribe_channels(ws)
                self._last_ws_message_sent_timestamp = self._time()
                while True:
                    try:
                        seconds_until_next_ping = (CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL -
                                                   (self._time() - self._last_ws_message_sent_timestamp))
                        await asyncio.wait_for(
                            self._process_ws_messages(ws=ws, output=output), timeout=seconds_until_next_ping)
                    except asyncio.TimeoutError:
                        ping_time = self._time()
                        payload = {
                            "ping": int(ping_time * 1e3)
                        }
                        ping_request = WSJSONRequest(payload=payload)
                        await ws.send(request=ping_request)
                        self._last_ws_message_sent_timestamp = ping_time
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error while listening to user stream. Retrying after 5 seconds...")
            finally:
                # Make sure no background task is leaked.
                ws and await ws.disconnect()
                await self._sleep(5)

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            trade_payload = {
                "id": "usertrade",
                "dataType": "spot.executionReport"
            }
            subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=trade_payload)

            balance_payload = {
                "id": "userbalance",
                "dataType": "ACCOUNT_UPDATE"
            }
            subscribe_balance_request: WSJSONRequest = WSJSONRequest(payload=balance_payload)

            await ws.send(subscribe_trade_request)
            await ws.send(subscribe_balance_request)

            self.logger().info("Subscribed to private channel...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                "Unexpected error occurred subscribing to order book trading and delta streams...",
                exc_info=True
            )
            raise

    # async def _authenticate_connection(self, ws: WSAssistant):
    #     """
    #     Sends the authentication message.
    #     :param ws: the websocket assistant used to connect to the exchange
    #     """
    #     auth_message: WSJSONRequest = WSJSONRequest(payload=self._auth.generate_ws_authentication_message())
    #     await ws.send(auth_message)

    async def _process_ws_messages(self, ws: WSAssistant, output: asyncio.Queue):
        # self.logger().info('process ws msgs')
        self._last_recv_time = self._time()
        async for ws_response in ws.iter_messages():
            data = utils.decompress_ws_message(ws_response.data)
            if data.get("e") == "ACCOUNT_UPDATE":
                output.put_nowait(data)
            elif (data.get("dataType") == "spot.executionReport"):
                output.put_nowait(data)
            # if isinstance(data, list):
            #     for message in data:
            #         if message["e"] in ["executionReport", "outboundAccountInfo"]:
            #             output.put_nowait(message)
            # elif data.get("auth") == "fail":
            #     raise IOError("Private channel authentication failed.")

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant

    def _time(self):
        return time.time()

    async def _get_listen_key(self):
        rest_assistant = await self._api_factory.get_rest_assistant()
        try:
            data = await rest_assistant.execute_request(
                url=web_utils.rest_url(path_url=CONSTANTS.USER_STREAM_PATH_URL, domain=self._domain),
                method=RESTMethod.POST,
                throttler_limit_id=CONSTANTS.USER_STREAM_PATH_URL,
                headers=self._auth.header_for_authentication()
            )
        except asyncio.CancelledError:
            raise
        except Exception as exception:
            raise IOError(f"Error fetching user stream listen key. Error: {exception}")

        return data["listenKey"]

    async def _ping_listen_key(self) -> bool:
        rest_assistant = await self._api_factory.get_rest_assistant()
        # self.logger().info("start renew listen key")
        try:
            data = await rest_assistant.execute_request(
                url=web_utils.rest_url(path_url=CONSTANTS.USER_STREAM_PATH_URL, domain=self._domain),
                params={"listenKey": self._current_listen_key},
                method=RESTMethod.PUT,
                return_err=True,
                throttler_limit_id=CONSTANTS.USER_STREAM_PATH_URL
            )
            self.logger().info(data)

        except asyncio.CancelledError:
            raise
        except Exception as exception:
            self.logger().warning(f"Failed to refresh the listen key {self._current_listen_key}: {exception}")
            return False

        return True

    async def _manage_listen_key_task_loop(self):
        try:
            while True:
                now = int(time.time())
                if self._current_listen_key is None:
                    self._current_listen_key = await self._get_listen_key()
                    self.logger().info(f"Successfully obtained listen key {self._current_listen_key}")
                    self._listen_key_initialized_event.set()
                    self._last_listen_key_ping_ts = int(time.time())

                if now - self._last_listen_key_ping_ts >= self.LISTEN_KEY_KEEP_ALIVE_INTERVAL:
                    success: bool = await self._ping_listen_key()
                    if not success:
                        self.logger().error("Error occurred renewing listen key ...")
                        break
                    else:
                        self.logger().info(f"Refreshed listen key {self._current_listen_key}.")
                        self._last_listen_key_ping_ts = int(time.time())
                else:
                    await self._sleep(self.LISTEN_KEY_KEEP_ALIVE_INTERVAL)
        finally:
            self._current_listen_key = None
            self._listen_key_initialized_event.clear()

    async def _connected_websocket_assistant(self) -> WSAssistant:
        """
        Creates an instance of WSAssistant connected to the exchange
        """
        self._manage_listen_key_task = safe_ensure_future(self._manage_listen_key_task_loop())
        await self._listen_key_initialized_event.wait()

        ws: WSAssistant = await self._get_ws_assistant()
        web_utils.wss_url(path_url=CONSTANTS.USER_STREAM_PATH_URL, domain=self._domain),
        url = f"{CONSTANTS.WSS_PRIVATE_URL[self._domain]}?listenKey={self._current_listen_key}"
        await ws.connect(ws_url=url, ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _on_user_stream_interruption(self, websocket_assistant: Optional[WSAssistant]):
        await super()._on_user_stream_interruption(websocket_assistant=websocket_assistant)
        self._manage_listen_key_task and self._manage_listen_key_task.cancel()
        self._current_listen_key = None
        self._listen_key_initialized_event.clear()
        await self._sleep(5)



from decimal import Decimal

from hummingbot.client.config.fee_overrides_config_map import fee_overrides_config_map
from hummingbot.client.settings import AllConnectorSettings
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeSchema


class TradeFeeSchemaLoader:
    """
    Utility class that contains the requried logic to load fee schemas applying any override the user
    might have configured.
    """

    @classmethod
    def configured_schema_for_exchange(cls, exchange_name: str) -> TradeFeeSchema:
        if exchange_name not in AllConnectorSettings.get_connector_settings():
            raise Exception(f"Invalid connector. {exchange_name} does not exist in AllConnectorSettings")
        trade_fee_schema = AllConnectorSettings.get_connector_settings()[exchange_name].trade_fee_schema
        trade_fee_schema = cls._superimpose_overrides(exchange_name, trade_fee_schema)
        return trade_fee_schema

    @classmethod
    def _superimpose_overrides(cls, exchange: str, trade_fee_schema: TradeFeeSchema):
        percent_fee_token_config = fee_overrides_config_map.get(f"{exchange}_percent_fee_token")
        trade_fee_schema.percent_fee_token = (
            percent_fee_token_config.value if percent_fee_token_config else None
        ) or trade_fee_schema.percent_fee_token

        maker_percent_fee_config = fee_overrides_config_map.get(f"{exchange}_maker_percent_fee")
        if maker_percent_fee_config and maker_percent_fee_config.value is not None:
            trade_fee_schema.maker_percent_fee_decimal = maker_percent_fee_config.value / Decimal("100")

        taker_percent_fee_config = fee_overrides_config_map.get(f"{exchange}_taker_percent_fee")
        if taker_percent_fee_config and taker_percent_fee_config.value is not None:
            trade_fee_schema.taker_percent_fee_decimal = taker_percent_fee_config.value / Decimal("100")

        buy_percent_fee_config = fee_overrides_config_map.get(f"{exchange}_buy_percent_fee_deducted_from_returns")
        if buy_percent_fee_config and buy_percent_fee_config.value is not None:
            trade_fee_schema.buy_percent_fee_deducted_from_returns = buy_percent_fee_config.value

        maker_fixed_fees_config = fee_overrides_config_map.get(f"{exchange}_maker_fixed_fees")
        trade_fee_schema.maker_fixed_fees = (
            maker_fixed_fees_config.value if maker_fixed_fees_config else None
        ) or trade_fee_schema.maker_fixed_fees
        trade_fee_schema.maker_fixed_fees = [
            TokenAmount(*maker_fixed_fee)
            for maker_fixed_fee in trade_fee_schema.maker_fixed_fees
        ]

        taker_fixed_fees_config = fee_overrides_config_map.get(f"{exchange}_taker_fixed_fees")
        trade_fee_schema.taker_fixed_fees = (
            taker_fixed_fees_config.value if taker_fixed_fees_config else None
        ) or trade_fee_schema.taker_fixed_fees
        trade_fee_schema.taker_fixed_fees = [
            TokenAmount(*taker_fixed_fee)
            for taker_fixed_fee in trade_fee_schema.taker_fixed_fees
        ]
        trade_fee_schema.validate_schema()
        return trade_fee_schema



from eip712_structs import Address, Array, Bytes, EIP712Struct, Int, String, Uint


class EIP712Domain(EIP712Struct):
    name = String()
    version = String()
    chainId = Uint(256)
    verifyingContract = Address()


# https://vertex-protocol.gitbook.io/docs/developer-resources/api/websocket-rest-api/executes/place-order
class Order(EIP712Struct):
    sender = Bytes(32)
    priceX18 = Int(128)
    amount = Int(128)
    expiration = Uint(64)
    nonce = Uint(64)


# https://vertex-protocol.gitbook.io/docs/developer-resources/api/websocket-rest-api/executes/cancel-orders
class Cancellation(EIP712Struct):
    sender = Bytes(32)
    productIds = Array(Uint(32))
    digests = Array(Bytes(32))
    nonce = Uint(64)


# https://vertex-protocol.gitbook.io/docs/developer-resources/api/websocket-rest-api/executes/cancel-product-orders
class CancellationProducts(EIP712Struct):
    sender = Bytes(32)
    productIds = Array(Uint(32))
    nonce = Uint(64)



import logging
from typing import Any, Dict, List, Optional

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.data_feed.candles_feed.bitmart_perpetual_candles import constants as CONSTANTS
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.logger import HummingbotLogger


class BitmartPerpetualCandles(CandlesBase):
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 150):
        super().__init__(trading_pair, interval, max_records)
        self.contract_size = None
        self.ws_interval = {
            "1m": "1m",
            "5m": "5m",
            "15m": "15m",
            "30m": "30m",
            "1h": "1H",
            "2h": "2H",
            "4h": "4H",
            "12h": "12H",
            "1d": "1D",
            "1w": "1W",
        }

    async def initialize_exchange_data(self):
        await self.get_exchange_trading_pair_contract_size()

    async def get_exchange_trading_pair_contract_size(self):
        contract_size = None
        rest_assistant = await self._api_factory.get_rest_assistant()
        response = await rest_assistant.execute_request(
            url=self.rest_url + CONSTANTS.CONTRACT_INFO_URL.format(contract=self._ex_trading_pair),
            throttler_limit_id=CONSTANTS.CONTRACT_INFO_URL
        )
        if response["code"] == 1000:
            symbols_data = response["data"].get("symbols")
            if len(symbols_data) > 0:
                contract_size = float(symbols_data[0]["contract_size"])
                self.contract_size = contract_size
        return contract_size

    @property
    def name(self):
        return f"bitmart_perpetual_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        return CONSTANTS.WSS_URL

    @property
    def health_check_url(self):
        return self.rest_url + CONSTANTS.HEALTH_CHECK_ENDPOINT

    @property
    def candles_url(self):
        return self.rest_url + CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    @property
    def is_linear(self):
        return "USDT" in self._trading_pair

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        await rest_assistant.execute_request(url=self.health_check_url,
                                             throttler_limit_id=CONSTANTS.HEALTH_CHECK_ENDPOINT)
        return NetworkStatus.CONNECTED

    def get_exchange_trading_pair(self, trading_pair):
        return trading_pair.replace("-", "")

    @property
    def _is_first_candle_not_included_in_rest_request(self):
        return False

    @property
    def _is_last_candle_not_included_in_rest_request(self):
        return False

    def _get_rest_candles_params(self,
                                 start_time: Optional[int] = None,
                                 end_time: Optional[int] = None,
                                 limit: Optional[int] = CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST) -> dict:
        """
        For API documentation, please refer to:
        https://developer-pro.bitmart.com/en/futuresv2/#get-k-line

        start_time and end_time must be used at the same time.
        """
        params = {
            "symbol": self._ex_trading_pair,
            "step": CONSTANTS.INTERVALS[self.interval],
        }
        if start_time:
            params["start_time"] = start_time
        if end_time:
            params["end_time"] = end_time
        return params

    def _parse_rest_candles(self, data: dict, end_time: Optional[int] = None) -> List[List[float]]:
        if data is not None and data.get("data") is not None:
            candles = data.get("data")
            if len(candles) > 0:
                return [[
                    self.ensure_timestamp_in_seconds(row["timestamp"]),
                    row["open_price"],
                    row["high_price"],
                    row["low_price"],
                    row["close_price"],
                    float(row["volume"]) * self.contract_size,
                    0.,
                    0.,
                    0.,
                    0.] for row in candles]
        return []

    def ws_subscription_payload(self):
        interval = self.ws_interval[self.interval]
        channel = f"futures/klineBin{interval}"
        args = [f"{channel}:{self._ex_trading_pair}"]
        payload = {
            "action": "subscribe",
            "args": args,
        }
        return payload

    def _parse_websocket_message(self, data):
        candles_row_dict: Dict[str, Any] = {}
        if data is not None and data.get("data") is not None:
            candle = data["data"]["items"][0]
            candles_row_dict["timestamp"] = self.ensure_timestamp_in_seconds(candle["ts"])
            candles_row_dict["open"] = candle["o"]
            candles_row_dict["low"] = candle["l"]
            candles_row_dict["high"] = candle["h"]
            candles_row_dict["close"] = candle["c"]
            candles_row_dict["volume"] = float(candle["v"]) * self.contract_size
            candles_row_dict["quote_asset_volume"] = 0.
            candles_row_dict["n_trades"] = 0.
            candles_row_dict["taker_buy_base_volume"] = 0.
            candles_row_dict["taker_buy_quote_volume"] = 0.
            return candles_row_dict



from sqlalchemy import Column, Integer, Text

from hummingbot.model.db_migration.base_transformation import DatabaseTransformation
from hummingbot.model.decimal_type_decorator import SqliteDecimal
from hummingbot.model.sql_connection_manager import SQLConnectionManager


class AddExchangeOrderIdColumnToOrders(DatabaseTransformation):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def apply(self, db_handle: SQLConnectionManager) -> SQLConnectionManager:
        exchange_order_id_column = Column("exchange_order_id", Text, nullable=True)
        self.add_column(db_handle.engine, "Order", exchange_order_id_column, dry_run=False)
        return db_handle

    @property
    def name(self):
        return "AddExchangeOrderIdColumnToOrders"

    @property
    def to_version(self):
        return 20210118


class AddLeverageAndPositionColumns(DatabaseTransformation):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def apply(self, db_handle: SQLConnectionManager) -> SQLConnectionManager:
        leverage_column = Column("leverage", Integer, nullable=True)
        position_column = Column("position", Text, nullable=True)
        self.add_column(db_handle.engine, "Order", leverage_column, dry_run=False)
        self.add_column(db_handle.engine, "Order", position_column, dry_run=False)
        self.add_column(db_handle.engine, "TradeFill", leverage_column, dry_run=False)
        self.add_column(db_handle.engine, "TradeFill", position_column, dry_run=False)
        return db_handle

    @property
    def name(self):
        return "AddLeverageAndPositionColumns"

    @property
    def to_version(self):
        return 20210119


class ConvertPriceAndAmountColumnsToBigint(DatabaseTransformation):
    order_migration_queries = [
        ('create table Order_dg_tmp'
         '(	id TEXT not null'
         '		primary key,'
         '	config_file_path TEXT not null,'
         '	strategy TEXT not null,'
         '	market TEXT not null,'
         '	symbol TEXT not null,'
         '	base_asset TEXT not null,'
         '	quote_asset TEXT not null,'
         '	creation_timestamp BIGINT not null,'
         '	order_type TEXT not null,'
         '	amount BIGINT not null,'
         '	leverage INTEGER not null,'
         '	price FLOAT not null,'
         '	last_status TEXT not null,'
         '	last_update_timestamp BIGINT not null,'
         '	exchange_order_id TEXT,'
         '	position TEXT);'),
        ('insert into Order_dg_tmp(id, config_file_path, strategy, market, symbol, base_asset, '
         'quote_asset, creation_timestamp, order_type, amount, leverage, price, last_status, '
         'last_update_timestamp, exchange_order_id, position) '
         'select id, config_file_path, strategy, market, symbol, base_asset, quote_asset, '
         'creation_timestamp, order_type, CAST(amount * 1000000 AS INTEGER), leverage, '
         'CAST(price * 1000000 AS INTEGER), last_status, last_update_timestamp, exchange_order_id, '
         'position from "Order";'),
        'drop table "Order";',
        'alter table Order_dg_tmp rename to "Order";',
        'create index o_config_timestamp_index on "Order" (config_file_path, creation_timestamp);',
        'create index o_market_base_asset_timestamp_index on "Order" (market, base_asset, creation_timestamp);',
        'create index o_market_quote_asset_timestamp_index on "Order" (market, quote_asset, creation_timestamp);',
        'create index o_market_trading_pair_timestamp_index on "Order" (market, symbol, creation_timestamp);'
    ]

    trade_fill_migration_queries = [
        ('create table TradeFill_dg_tmp'
         '(	config_file_path TEXT not null,'
         '	strategy TEXT not null,'
         '	market TEXT not null,'
         '	symbol TEXT not null,'
         '	base_asset TEXT not null,'
         '	quote_asset TEXT not null,'
         '	timestamp BIGINT not null,'
         '	order_id TEXT not null'
         '		references "Order",'
         '	trade_type TEXT not null,'
         '	order_type TEXT not null,'
         '	price BIGINT not null,'
         '	amount FLOAT not null,'
         '	leverage INTEGER not null,'
         '	trade_fee JSON not null,'
         '	exchange_trade_id TEXT not null,'
         '	position TEXT,'
         '	constraint TradeFill_pk'
         '	primary key (market, order_id, exchange_trade_id)'
         ');'),
        ('insert into TradeFill_dg_tmp(config_file_path, strategy, market, symbol, base_asset, '
         'quote_asset, timestamp, order_id, trade_type, order_type, price, amount, leverage, '
         'trade_fee, exchange_trade_id, position) '
         'select config_file_path, strategy, market, symbol, base_asset, quote_asset, timestamp, '
         'order_id, trade_type, order_type, CAST(price * 1000000 AS INTEGER), '
         "CAST(amount * 1000000 AS INTEGER), leverage, trade_fee, exchange_trade_id || '_' || id, position "
         'from TradeFill;'),
        'drop table TradeFill;',
        'alter table TradeFill_dg_tmp rename to TradeFill;',
        'create index tf_config_timestamp_index on TradeFill (config_file_path, timestamp);',
        'create index tf_market_base_asset_timestamp_index on TradeFill (market, base_asset, timestamp);',
        'create index tf_market_quote_asset_timestamp_index on TradeFill (market, quote_asset, timestamp);',
        'create index tf_market_trading_pair_timestamp_index on TradeFill (market, symbol, timestamp);'
    ]

    def apply(self, db_handle: SQLConnectionManager) -> SQLConnectionManager:
        for query in self.order_migration_queries:
            db_handle.engine.execute(query)
        for query in self.trade_fill_migration_queries:
            db_handle.engine.execute(query)
        return db_handle

    @property
    def name(self):
        return "ConvertPriceAndAmountColumnsToBigint"

    @property
    def to_version(self):
        return 20220130


class AddTradeFeeInQuote(DatabaseTransformation):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def apply(self, db_handle: SQLConnectionManager) -> SQLConnectionManager:
        trade_fee = Column("trade_fee_in_quote", SqliteDecimal(6), nullable=True)
        self.add_column(db_handle.engine, "TradeFill", trade_fee, dry_run=False)
        return db_handle

    @property
    def name(self):
        return "AddTradeFeeInQuote"

    @property
    def to_version(self):
        return 20230516



# distutils: language=c++
# distutils: sources=hummingbot/core/cpp/LimitOrder.cpp
import time
from decimal import Decimal
from typing import List

import pandas as pd
from cpython cimport PyObject
from libcpp.string cimport string

from hummingbot.core.data_type.common import PositionAction, OrderType
from hummingbot.core.event.events import LimitOrderStatus

cdef class LimitOrder:
    """
    A Python wrapper class on C++ LimitOrder. This data class is used to store order information and it is passed around
    between connectors and strategies. It is also used in HummingSim for back testing as well.
    """
    @classmethod
    def to_pandas(cls, limit_orders: List[LimitOrder], mid_price: float = 0.0, hanging_ids: List[str] = None,
                  end_time_order_age: int = 0) \
            -> pd.DataFrame:
        """
        Creates a dataframe for displaying current active orders
        :param limit_orders: A list of current active LimitOrder from a single market
        :param mid_price: The mid price (between best bid and best ask) of the market
        :param hanging_ids: A list of hanging order ids if applicable
        :param end_time_order_age: The end time for order age calculation, if unspecified the current time is used.
        :return: A pandas data frame object
        """
        cdef:
            list buys = [o for o in limit_orders if o.is_buy]
            list sells = [o for o in limit_orders if not o.is_buy]
        buys.sort(key=lambda x: x.price, reverse=True)
        sells.sort(key=lambda x: x.price, reverse=True)
        cdef:
            list orders = []
            list columns = ["Order ID", "Type", "Price", "Spread", "Amount", "Age", "Hang"]
            list data = []
            str order_id_txt, type_txt, spread_txt, age_txt, hang_txt
            double price, quantity
            long long age_seconds
            long long now_timestamp = int(time.time() * 1e6) if end_time_order_age == 0 else end_time_order_age
        sells.extend(buys)
        for order in sells:
            order_id_txt = order.client_order_id if len(order.client_order_id) <= 7 else f"...{order.client_order_id[-4:]}"
            type_txt = "buy" if order.is_buy else "sell"
            price = float(order.price)
            spread_txt = f"{(0 if mid_price == 0 else abs(float(order.price) - mid_price) / mid_price):.2%}"
            quantity = float(order.quantity)
            age_txt = "n/a"
            age_seconds = order.age_til(now_timestamp)
            if age_seconds >= 0:
                age_txt = pd.Timestamp(age_seconds, unit='s', tz='UTC').strftime('%H:%M:%S')
            hang_txt = "n/a" if hanging_ids is None else ("yes" if order.client_order_id in hanging_ids else "no")
            data.append([order_id_txt, type_txt, price, spread_txt, quantity, age_txt, hang_txt])
        return pd.DataFrame(data=data, columns=columns)

    def __init__(self,
                 client_order_id: str,
                 trading_pair: str,
                 is_buy: bool,
                 base_currency: str,
                 quote_currency: str,
                 price: Decimal,
                 quantity: Decimal,
                 filled_quantity: Decimal = Decimal("NaN"),
                 creation_timestamp: int = 0,
                 status: LimitOrderStatus = LimitOrderStatus.UNKNOWN,
                 position: PositionAction = PositionAction.NIL):
        cdef:
            string cpp_client_order_id = client_order_id.encode("utf8")
            string cpp_trading_pair = trading_pair.encode("utf8")
            string cpp_base_currency = base_currency.encode("utf8")
            string cpp_quote_currency = quote_currency.encode("utf8")
            string cpp_position = position.value.encode("utf8")
        self._cpp_limit_order = CPPLimitOrder(cpp_client_order_id,
                                              cpp_trading_pair,
                                              is_buy,
                                              cpp_base_currency,
                                              cpp_quote_currency,
                                              <PyObject *> price,
                                              <PyObject *> quantity,
                                              <PyObject *> filled_quantity,
                                              creation_timestamp,
                                              status.value,
                                              cpp_position)

    @property
    def client_order_id(self) -> str:
        cdef:
            string cpp_client_order_id = self._cpp_limit_order.getClientOrderID()
            str retval = cpp_client_order_id.decode("utf8")
        return retval

    @property
    def trading_pair(self) -> str:
        cdef:
            string cpp_trading_pair = self._cpp_limit_order.getTradingPair()
            str retval = cpp_trading_pair.decode("utf8")
        return retval

    @property
    def is_buy(self) -> bool:
        return self._cpp_limit_order.getIsBuy()

    @property
    def base_currency(self) -> str:
        cdef:
            string cpp_base_currency = self._cpp_limit_order.getBaseCurrency()
            str retval = cpp_base_currency.decode("utf8")
        return retval

    @property
    def quote_currency(self) -> str:
        cdef:
            string cpp_quote_currency = self._cpp_limit_order.getQuoteCurrency()
            str retval = cpp_quote_currency.decode("utf8")
        return retval

    @property
    def price(self) -> Decimal:
        return <object>(self._cpp_limit_order.getPrice())

    @property
    def quantity(self) -> Decimal:
        return <object>(self._cpp_limit_order.getQuantity())

    @property
    def filled_quantity(self) -> Decimal:
        return <object>(self._cpp_limit_order.getFilledQuantity())

    @property
    def creation_timestamp(self) -> int:
        return self._cpp_limit_order.getCreationTimestamp()

    @property
    def status(self) -> LimitOrderStatus:
        return LimitOrderStatus(self._cpp_limit_order.getStatus())

    @property
    def position(self) -> PositionAction:
        cdef:
            string cpp_position = self._cpp_limit_order.getPosition()
            str retval = cpp_position.decode("utf8")
        return PositionAction(retval)

    cdef long long c_age_til(self, long long end_timestamp):
        """
        Calculates and returns age of the order since it was created til end_timestamp in seconds
        :param end_timestamp: The end timestamp
        :return: The age of the order in seconds
        """
        cdef long long start_timestamp = 0
        if self.creation_timestamp > 0:
            start_timestamp = self.creation_timestamp
        elif len(self.client_order_id) > 16 and self.client_order_id[-16:].isnumeric():
            start_timestamp = int(self.client_order_id[-16:])
        if 0 < start_timestamp < end_timestamp:
            return <long long>((end_timestamp - start_timestamp) / 1e6)
        else:
            return -1

    cdef long long c_age(self):
        """
        Calculates and returns age of the order since it was created til now.
        """
        return self.c_age_til(int(time.time() * 1e6))

    def age(self) -> int:
        return self.c_age()

    def age_til(self, start_timestamp: int) -> int:
        return self.c_age_til(start_timestamp)

    def order_type(self) -> OrderType:
        return OrderType.LIMIT

    def copy_with_id(self, client_order_id: str):
        return LimitOrder(
            client_order_id=client_order_id,
            trading_pair=self.trading_pair,
            is_buy=self.is_buy,
            base_currency=self.base_currency,
            quote_currency=self.quote_currency,
            price=self.price,
            quantity=self.quantity,
            filled_quantity=self.filled_quantity,
            creation_timestamp=self.creation_timestamp,
            status=self.status,
            position=self.position,
        )

    def __repr__(self) -> str:
        return (f"LimitOrder('{self.client_order_id}', '{self.trading_pair}', {self.is_buy}, '{self.base_currency}', "
                f"'{self.quote_currency}', {self.price}, {self.quantity}, {self.filled_quantity}, "
                f"{self.creation_timestamp})")


cdef LimitOrder c_create_limit_order_from_cpp_limit_order(const CPPLimitOrder cpp_limit_order):
    cdef LimitOrder retval = LimitOrder.__new__(LimitOrder)
    retval._cpp_limit_order = cpp_limit_order
    return retval



import asyncio
import json
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.gate_io import gate_io_constants as CONSTANTS, gate_io_web_utils as web_utils
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.gate_io.gate_io_exchange import GateIoExchange


class GateIoAPIOrderBookDataSource(OrderBookTrackerDataSource):

    _logger: Optional[HummingbotLogger] = None
    _DYNAMIC_SUBSCRIBE_ID_START = 100
    _next_subscribe_id: int = _DYNAMIC_SUBSCRIBE_ID_START

    def __init__(self,
                 trading_pairs: List[str],
                 connector: 'GateIoExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        super().__init__(trading_pairs)
        self._connector = connector
        self._api_factory = api_factory
        self._trading_pairs: List[str] = trading_pairs

        self._message_queue: Dict[str, asyncio.Queue] = defaultdict(asyncio.Queue)

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_response: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_timestamp: float = self._time()
        snapshot_msg: OrderBookMessage = OrderBookMessage(
            OrderBookMessageType.SNAPSHOT,
            {
                "trading_pair": trading_pair,
                "update_id": snapshot_response["id"],
                "bids": snapshot_response["bids"],
                "asks": snapshot_response["asks"],
            },
            timestamp=snapshot_timestamp)
        return snapshot_msg

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.

        :param trading_pair: the trading pair for which the order book will be retrieved

        :return: the response from the exchange (JSON dictionary)
        """
        params = {
            "currency_pair": await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
            "with_id": json.dumps(True)
        }

        rest_assistant = await self._api_factory.get_rest_assistant()
        return await rest_assistant.execute_request(
            url=web_utils.public_rest_url(endpoint=CONSTANTS.ORDER_BOOK_PATH_URL),
            params=params,
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.ORDER_BOOK_PATH_URL,
        )

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        trade_data: Dict[str, Any] = raw_message["result"]
        trade_timestamp: float = float(trade_data["create_time_ms"]) * 1e-3
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(
            symbol=trade_data["currency_pair"])
        message_content = {
            "trading_pair": trading_pair,
            "trade_type": (float(TradeType.SELL.value)
                           if trade_data["side"] == "sell"
                           else float(TradeType.BUY.value)),
            "trade_id": trade_data["id"],
            "update_id": trade_timestamp,
            "price": trade_data["price"],
            "amount": trade_data["amount"],
        }
        trade_message: Optional[OrderBookMessage] = OrderBookMessage(
            message_type=OrderBookMessageType.TRADE,
            content=message_content,
            timestamp=trade_timestamp)

        message_queue.put_nowait(trade_message)

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        diff_data: [str, Any] = raw_message["result"]
        timestamp: float = (diff_data["t"]) * 1e-3
        update_id: int = diff_data["u"]

        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol=diff_data["s"])

        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": update_id,
            "first_update_id": diff_data["U"],
            "bids": diff_data["b"],
            "asks": diff_data["a"],
        }
        diff_message: OrderBookMessage = OrderBookMessage(
            OrderBookMessageType.DIFF,
            order_book_message_content,
            timestamp)

        message_queue.put_nowait(diff_message)

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.

        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            for trading_pair in self._trading_pairs:
                symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

                trades_payload = {
                    "time": int(self._time()),
                    "channel": CONSTANTS.TRADES_ENDPOINT_NAME,
                    "event": "subscribe",
                    "payload": [symbol]
                }
                subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=trades_payload)

                order_book_payload = {
                    "time": int(self._time()),
                    "channel": CONSTANTS.ORDERS_UPDATE_ENDPOINT_NAME,
                    "event": "subscribe",
                    "payload": [symbol, "100ms"]
                }
                subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=order_book_payload)

                await ws.send(subscribe_trade_request)
                await ws.send(subscribe_orderbook_request)

                self.logger().info("Subscribed to public order book and trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error("Unexpected error occurred subscribing to order book data streams.")
            raise

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if event_message.get("error") is not None:
            err_msg = event_message.get("error", {}).get("message", event_message.get("error"))
            raise IOError(f"Error event received from the server ({err_msg})")
        elif event_message.get("event") == "update":
            if event_message.get("channel") == CONSTANTS.ORDERS_UPDATE_ENDPOINT_NAME:
                channel = self._diff_messages_queue_key
            elif event_message.get("channel") == CONSTANTS.TRADES_ENDPOINT_NAME:
                channel = self._trade_messages_queue_key

        return channel

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=CONSTANTS.WS_URL, ping_timeout=CONSTANTS.PING_TIMEOUT)
        return ws

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """
        Subscribes to order book and trade channels for a single trading pair
        on the existing WebSocket connection.

        :param trading_pair: the trading pair to subscribe to
        :return: True if subscription was successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot subscribe to {trading_pair}: WebSocket not connected"
            )
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

            trades_payload = {
                "time": int(self._time()),
                "channel": CONSTANTS.TRADES_ENDPOINT_NAME,
                "event": "subscribe",
                "payload": [symbol]
            }
            subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=trades_payload)

            order_book_payload = {
                "time": int(self._time()),
                "channel": CONSTANTS.ORDERS_UPDATE_ENDPOINT_NAME,
                "event": "subscribe",
                "payload": [symbol, "100ms"]
            }
            subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=order_book_payload)

            await self._ws_assistant.send(subscribe_trade_request)
            await self._ws_assistant.send(subscribe_orderbook_request)

            self.add_trading_pair(trading_pair)
            self.logger().info(f"Subscribed to {trading_pair} order book and trade channels")
            return True

        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception(f"Error subscribing to {trading_pair}")
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """
        Unsubscribes from order book and trade channels for a single trading pair
        on the existing WebSocket connection.

        :param trading_pair: the trading pair to unsubscribe from
        :return: True if unsubscription was successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot unsubscribe from {trading_pair}: WebSocket not connected"
            )
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

            trades_payload = {
                "time": int(self._time()),
                "channel": CONSTANTS.TRADES_ENDPOINT_NAME,
                "event": "unsubscribe",
                "payload": [symbol]
            }
            unsubscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=trades_payload)

            order_book_payload = {
                "time": int(self._time()),
                "channel": CONSTANTS.ORDERS_UPDATE_ENDPOINT_NAME,
                "event": "unsubscribe",
                "payload": [symbol, "100ms"]
            }
            unsubscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=order_book_payload)

            await self._ws_assistant.send(unsubscribe_trade_request)
            await self._ws_assistant.send(unsubscribe_orderbook_request)

            self.remove_trading_pair(trading_pair)
            self.logger().info(f"Unsubscribed from {trading_pair} order book and trade channels")
            return True

        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception(f"Error unsubscribing from {trading_pair}")
            return False

    @classmethod
    def _get_next_subscribe_id(cls) -> int:
        """Returns the next subscription ID and increments the counter."""
        current_id = cls._next_subscribe_id
        cls._next_subscribe_id += 1
        return current_id
